from datetime import datetime

from dateutil.tz import gettz
from marshmallow import pre_load

from .._base import SystemBaseSchema


class UserBaseSchema(SystemBaseSchema):
    """
    Base schema for system users.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        started_ts = data.pop('started', None)
        if started_ts is not None:
            data['started'] = datetime.fromtimestamp(started_ts).replace(tzinfo=gettz())

        data['username'] = data.pop('name', data.pop('username', None))
        return data



from sqlalchemy import Column, Integer, Boolean, ForeignKey

from platypush.common.db import is_defined

from ._base import Entity


if not is_defined('device'):

    class Device(Entity):
        """
        Base class to model device entities.
        """

        __tablename__ = 'device'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )
        reachable = Column(Boolean, default=True)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



import base64

from marshmallow import EXCLUDE, fields, pre_dump
from marshmallow.schema import Schema


class QrcodeGeneratedSchema(Schema):
    """
    Schema for a QR code generation response.
    """

    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    text = fields.String(
        required=True,
        metadata={
            'description': 'Text content of the QR code',
            'example': 'https://platypush.tech',
        },
    )

    data = fields.String(
        metadata={
            'description': 'Base64-encoded content of the QR code',
            'example': 'iVBORw0KGgoAAAANSUhEUgAAAXIAAAFyAQAAAADAX2yk',
        }
    )

    format = fields.String(
        metadata={
            'description': 'Format of the QR code image',
            'example': 'png',
        },
    )

    image_file = fields.String(
        metadata={
            'description': 'Path to the generated QR code image file',
            'example': '/tmp/qr_code.png',
        },
    )


class QrcodeDecodedRectSchema(Schema):
    """
    Schema for a single QR code decoding result rectangle.
    """

    x = fields.Integer(
        required=True,
        metadata={
            'description': 'X coordinate of the rectangle in the image',
            'example': 0,
        },
    )

    y = fields.Integer(
        required=True,
        metadata={
            'description': 'Y coordinate of the rectangle in the image',
            'example': 0,
        },
    )

    width = fields.Integer(
        required=True,
        metadata={
            'description': 'Width of the rectangle',
            'example': 100,
        },
    )

    height = fields.Integer(
        required=True,
        metadata={
            'description': 'Height of the rectangle',
            'example': 100,
        },
    )


class QrcodeDecodedResultSchema(Schema):
    """
    Schema for a single QR code decoding result.
    """

    data = fields.String(
        required=True,
        metadata={
            'description': 'Decoded QRcode data, as a base64-encoded string if binary',
            'example': 'https://platypush.tech',
        },
    )

    type = fields.String(
        required=True,
        metadata={
            'description': (
                'Type of code that was decoded. Supports the types available under the '
                '`pyzbar.ZBarSymbol` class: '
                'https://github.com/NaturalHistoryMuseum/pyzbar/blob/master/pyzbar/wrapper.py#L43'
            ),
            'example': 'QRCODE',
        },
    )

    rect = fields.Nested(
        QrcodeDecodedRectSchema,
        required=True,
        metadata={
            'description': 'Rectangle in the image where the QR code was found',
        },
    )

    @pre_dump
    def pre_dump(self, data, **_):
        if hasattr(data, '_asdict'):
            data = data._asdict()

        try:
            data['data'] = data['data'].decode()
        except (ValueError, TypeError):
            data['data'] = base64.b64encode(data['data']).decode()

        return data


class QrcodeDecodedSchema(Schema):
    """
    Schema for a QR code decoding response.
    """

    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    results = fields.List(
        fields.Nested(QrcodeDecodedResultSchema),
        required=True,
        metadata={
            'description': 'Decoded QR code results',
        },
    )

    image_file = fields.String(
        metadata={
            'description': 'Path to the image file that was decoded',
            'example': '/tmp/qr_code.png',
        },
    )



from abc import ABC, abstractmethod
from typing import Iterable

from platypush.entities import Entity
from platypush.entities.bluetooth import BluetoothDevice
from platypush.plugins.bluetooth._manager import BaseBluetoothManager


class BaseBluetoothPlugin(ABC):
    """
    Base class for Bluetooth plugins, like Switchbot or HID integrations.
    """

    def __init__(self, manager: BaseBluetoothManager):
        self._manager = manager

    @abstractmethod
    def supports_device(self, device: BluetoothDevice) -> bool:
        """
        Returns True if the given device matches this plugin.
        """

    @abstractmethod
    def _extract_entities(self, device: BluetoothDevice) -> Iterable[Entity]:
        """
        If :meth:`_matches_device` returns True, this method should return an
        iterable of entities that will be used to extend the existing device -
        like sensors, switches etc.
        """

    def extend_device(self, device: BluetoothDevice):
        """
        Extends the given device with entities extracted through
        :meth:`_extract_entities`, if :meth:`_matches_device` returns True.
        """
        if not self.supports_device(device):
            return

        device.children.extend(self._extract_entities(device))



from abc import ABC, abstractmethod
from typing import Dict, Iterable, List, Union

from .._model import Mail
from ._base import BaseMailPlugin


class MailInPlugin(BaseMailPlugin, ABC):
    """
    Base class for mail in plugins.
    """

    @abstractmethod
    def get_folders(self, **_) -> list:
        raise NotImplementedError()

    @abstractmethod
    def get_sub_folders(self, **_) -> list:
        raise NotImplementedError()

    @abstractmethod
    def search(
        self, criteria: Union[str, Iterable[str]], folder: str, **_
    ) -> List[Mail]:
        raise NotImplementedError()

    @abstractmethod
    def search_unseen_messages(self, folder: str) -> List[Mail]:
        raise NotImplementedError()

    @abstractmethod
    def search_flagged_messages(self, folder: str, **_) -> List[Mail]:
        raise NotImplementedError()

    @abstractmethod
    def search_starred_messages(self, folder: str, **_) -> List[Mail]:
        raise NotImplementedError()

    @abstractmethod
    def sort(
        self,
        folder: str,
        sort_criteria: Union[str, Iterable[str]],
        criteria: Union[str, Iterable[str]],
    ) -> list:
        raise NotImplementedError()

    @abstractmethod
    def get_messages(self, *ids, with_body: bool = True, **_) -> Dict[int, Mail]:
        raise NotImplementedError()

    def get_message(
        self, id, with_body: bool = True, **_  # pylint: disable=redefined-builtin
    ) -> Mail:
        msgs = self.get_messages(id, with_body=with_body)
        msg = msgs.get(id)
        assert msg, f"Message {id} not found"
        return msg

    @abstractmethod
    def create_folder(self, folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def rename_folder(self, old_name: str, new_name: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def delete_folder(self, folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def add_flags(
        self, messages: list, flags: Union[str, Iterable[str]], folder: str, **_
    ):
        raise NotImplementedError()

    @abstractmethod
    def set_flags(
        self, messages: list, flags: Union[str, Iterable[str]], folder: str, **_
    ):
        raise NotImplementedError()

    @abstractmethod
    def remove_flags(
        self, messages: list, flags: Union[str, Iterable[str]], folder: str, **_
    ):
        raise NotImplementedError()

    @abstractmethod
    def delete_messages(self, messages: list, folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def restore_messages(self, messages: list, folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def copy_messages(self, messages: list, dest_folder: str, source_folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def move_messages(self, messages: list, dest_folder: str, source_folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def expunge_messages(self, folder: str, messages: list, **_):
        raise NotImplementedError()

    @abstractmethod
    def flag_messages(self, messages: list, folder: str, **_):
        raise NotImplementedError()

    @abstractmethod
    def unflag_messages(self, messages: List[int], folder: str = 'INBOX', **_):
        raise NotImplementedError()

    def flag_message(self, message: int, folder: str, **_):
        return self.flag_messages([message], folder=folder)

    def unflag_message(self, message: int, folder: str, **_):
        return self.unflag_messages([message], folder=folder)


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional


@dataclass
class XmppConfig:
    """
    Data class that models the XMPP configuration shared across all submodules.
    """

    auto_accept_invites: bool = True
    """
    Whether or not to automatically accept invites to rooms and buddy lists.
    """

    restore_state: bool = True
    """
    Whether to restore the previous state of the joined rooms and subscriptions
    upon application restart.
    """

    state_file: Optional[str] = None
    """
    The path where the state of the client is persisted across sessions.
    """



import time
from typing import Optional

import todoist
import todoist.managers.items  # type: ignore
import websocket

from platypush.plugins import Plugin, action
from platypush.schemas.todoist import (
    TodoistCollaboratorSchema,
    TodoistFilterSchema,
    TodoistItemSchema,
    TodoistLiveNotificationSchema,
    TodoistNoteSchema,
    TodoistProjectNoteSchema,
    TodoistProjectSchema,
    TodoistUserSchema,
)


class TodoistPlugin(Plugin):
    """
    Todoist integration.

    You'll also need a Todoist token. You can get it `here
    <https://todoist.com/prefs/integrations>`_.
    """

    _sync_timeout = 60.0

    def __init__(self, api_token: str, **kwargs):
        """
        :param api_token: Todoist API token.
            You can get it `here <https://todoist.com/prefs/integrations>`_.
        """

        super().__init__(**kwargs)
        self.api_token = api_token
        self._api = None
        self._last_sync_time = None

        self._ws_url: Optional[str] = None
        self._ws: Optional[websocket.WebSocketApp] = None
        self._connected = False
        self._todoist_initialized = False

        self._items = {}
        self._event_handled = False

    def _get_api(self) -> todoist.TodoistAPI:  # type: ignore
        if not self._api:
            self._api = todoist.TodoistAPI(self.api_token)  # type: ignore

        if (
            not self._last_sync_time
            or time.time() - self._last_sync_time > self._sync_timeout
        ):
            self._api.sync()

        return self._api

    @action
    def get_user(self):
        """
        Get logged user info.

        :return: .. schema:: todoist.TodoistUserSchema
        """
        return TodoistUserSchema().dump(self._get_api().state['user'])

    @action
    def get_projects(self):
        """
        Get list of Todoist projects.

        :return: .. schema:: todoist.TodoistProjectSchema(many=True)
        """
        return TodoistProjectSchema().dump(self._get_api().state['projects'], many=True)

    @action
    def get_items(self):
        """
        Get list of Todoist projects.

        :return .. schema:: todoist.TodoistItemSchema(many=True)
        """
        return TodoistItemSchema().dump(self._get_api().state['items'], many=True)

    @action
    def get_filters(self):
        """
        Get list of Todoist filters.

        :return: .. schema:: todoist.TodoistFilterSchema(many=True)
        """
        return TodoistFilterSchema().dump(self._get_api().state['filters'], many=True)

    @action
    def get_live_notifications(self):
        """
        Get list of Todoist live notifications.

        :return: .. schema:: todoist.TodoistLiveNotificationSchema(many=True)
        """
        return TodoistLiveNotificationSchema().dump(
            self._get_api().state['live_notifications'], many=True
        )

    @action
    def get_collaborators(self):
        """
        Get list of collaborators.

        :return: .. schema:: todoist.TodoistCollaboratorSchema(many=True)
        """
        return TodoistCollaboratorSchema().dump(
            self._get_api().state['collaborators'], many=True
        )

    @action
    def get_notes(self):
        """
        Get list of Todoist notes.

        :return: .. schema:: todoist.TodoistNoteSchema(many=True)
        """
        return TodoistNoteSchema().dump(self._get_api().state['notes'], many=True)

    @action
    def get_project_notes(self):
        """
        Get list of Todoist project notes.

        :return: .. schema:: todoist.TodoistProjectNoteSchema(many=True)
        """
        return TodoistProjectNoteSchema().dump(
            self._get_api().state['project_notes'], many=True
        )

    @action
    def add_item(self, content: str, project_id: Optional[int] = None, **kwargs):
        """
        Add a new item.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        item = manager.add(content, project_id=project_id, **kwargs)
        api.commit()
        return item.data

    @action
    def delete_item(self, item_id: int):
        """
        Delete an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.delete(item_id)
        api.commit()

    @action
    def update_item(self, item_id: int, **kwargs):
        """
        Update an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.update(item_id, **kwargs)
        api.commit()

    @action
    def complete_item(self, item_id: int):
        """
        Mark an item as done.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.complete(item_id)
        api.commit()

    @action
    def uncomplete_item(self, item_id: int):
        """
        Mark an item as not done.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.uncomplete(item_id)
        api.commit()

    @action
    def archive(self, item_id: int):
        """
        Archive an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.archive(item_id)
        api.commit()

    @action
    def unarchive(self, item_id: int):
        """
        Un-archive an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.unarchive(item_id)
        api.commit()

    @action
    def sync(self):
        """
        Sync/update info with the remote server.
        """
        api = self._get_api()
        api.sync()


# vim:sw=4:ts=4:et:



import functools
import mimetypes
import os

from platypush.utils import get_mime_type

from . import MediaHandler


class FileHandler(MediaHandler):
    """
    Handler for local media files.
    """

    prefix_handlers = ['file://']

    def __init__(self, source, *args, **kwargs):
        super().__init__(source, *args, **kwargs)

        self.path = os.path.abspath(
            os.path.expanduser(self.source[len(self._matched_handler) :])
        )
        self.filename = self.path.split('/')[-1]

        if not os.path.isfile(self.path):
            raise FileNotFoundError(self.path)

        self.mime_type = get_mime_type(source)
        assert self.mime_type, f'Could not detect mime type for {source}'
        assert (
            self.mime_type[:5] in ['video', 'audio', 'image']
            or self.mime_type == 'application/octet-stream'
        ), f'{source} is not a valid media file (detected format: {self.mime_type})'

        self.extension = mimetypes.guess_extension(self.mime_type)
        if self.url and self.extension:
            self.url += self.extension
        self.content_length = os.path.getsize(self.path)

    def get_data(self, from_bytes=None, to_bytes=None, chunk_size=None):
        if from_bytes is None:
            from_bytes = 0
        if to_bytes is None:
            to_bytes = os.path.getsize(self.path)
        if chunk_size is None:
            chunk_size = os.path.getsize(self.path) - from_bytes

        with open(self.path, 'rb') as f:
            f.seek(from_bytes)
            for chunk in iter(
                functools.partial(f.read, min(to_bytes - from_bytes, chunk_size)), b''
            ):
                yield chunk


# vim:sw=4:ts=4:et:



import inspect
import textwrap as tw
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    get_type_hints,
)

from .._model.argument import Argument
from .._model.returns import ReturnValue
from .state import ParseState


@dataclass
class ParseContext:
    """
    Runtime parsing context.
    """

    obj: Callable
    state: ParseState = ParseState.DOC
    cur_param: Optional[str] = None
    doc: Optional[str] = None
    returns: ReturnValue = field(default_factory=ReturnValue)
    parsed_params: dict[str, Argument] = field(default_factory=dict)

    def __post_init__(self):
        """
        Initialize the return type and parameters from the function annotations.
        """

        # If we're dealing with a wrapped @action, then unwrap the underlying function
        if hasattr(self.obj, "wrapped"):
            self.obj = self.obj.wrapped

        # Initialize the return type from the annotations
        annotations = getattr(self.obj, "__annotations__", {})
        if annotations:
            self.returns.type = annotations.get("return")

        # Initialize the parameters from the signature
        defaults = self.spec.defaults or ()
        defaults = ((Any,) * (len(self.param_names) - len(defaults or ()))) + defaults
        self.parsed_params = {
            name: Argument(
                name=name,
                type=self.param_types.get(name),
                default=default if default is not Any else None,
                required=default is Any,
            )
            for name, default in zip(self.param_names, defaults)
        }

        # Update the parameters with the keyword-only arguments
        self.parsed_params.update(
            {
                arg: Argument(
                    name=arg,
                    type=self.param_types.get(arg),
                    default=(self.spec.kwonlydefaults or {}).get(arg),
                    required=arg not in (self.spec.kwonlydefaults or {}),
                )
                for arg in self.spec.kwonlyargs
            }
        )

    @property
    def spec(self) -> inspect.FullArgSpec:
        return inspect.getfullargspec(self.obj)

    @property
    def param_names(self) -> List[str]:
        return list(self.spec.args[1:])

    @property
    def param_defaults(self) -> Tuple[Any]:
        defaults = self.spec.defaults or ()
        return ((Any,) * (len(self.spec.args[1:]) - len(defaults))) + defaults

    @property
    def param_types(self) -> dict[str, Type]:
        return get_type_hints(self.obj)

    @property
    def doc_lines(self) -> Iterable[str]:
        return tw.dedent(inspect.getdoc(self.obj) or "").split("\n")



from typing import Optional, Collection, Mapping

from platypush.message.event import Event


class NotificationEvent(Event):
    """
    Event triggered when a message/notification is received on a subscribed
    channel.
    """

    def __init__(
        self,
        *args,
        id: str,
        topic: str,
        message: str,
        title: Optional[str] = None,
        priority: Optional[int] = None,
        time: Optional[int] = None,
        attachment: Optional[Mapping] = None,
        actions: Optional[Collection[Mapping]] = None,
        tags: Optional[Collection[str]] = None,
        url: Optional[str] = None,
        **kwargs
    ):
        """
        :param id: Message ID.
        :param topic: The topic where the message was received.
        :param message: Message body.
        :param title: Message title.
        :param priority: Message priority.
        :param time: Message UNIX timestamp.
        :param tags: Notification tags.
        :param url: URL spawned when the notification is clicked.
        :param actions: List of actions associated to the notification.
            Example:

            .. code-block:: json

                [
                    {
                        "action": "view",
                        "label": "Open portal",
                        "url": "https://home.nest.com/",
                        "clear": true
                    },
                    {
                        "action": "http",
                        "label": "Turn down",
                        "url": "https://api.nest.com/",
                        "method": "PUT",
                        "headers": {
                            "Authorization": "Bearer abcdef..."
                        },
                        "body": "{\\"temperature\\": 65}"
                    },
                    {
                        "action": "broadcast",
                        "label": "Take picture",
                        "intent": "com.myapp.TAKE_PICTURE_INTENT",
                        "extras": {
                            "camera": "front"
                        }
                    }
                ]

        :param attachment: Attachment metadata. Example:

            .. code-block:: json

                {
                  "name": "image.jpg",
                  "type": "image/jpeg",
                  "size": 30017,
                  "expires": 1654144935,
                  "url": "https://ntfy.example.com/file/01234abcd.jpg"
                }

        """
        super().__init__(
            *args,
            id=id,
            topic=topic,
            message=message,
            title=title,
            priority=priority,
            time=time,
            tags=tags,
            attachment=attachment,
            actions=actions,
            url=url,
            **kwargs
        )



from marshmallow import Schema, fields

from platypush.schemas import DateTime


class WallabagSchema(Schema):
    pass


class WallabagAnnotationSchema(WallabagSchema):
    id = fields.Integer(
        required=True,
        dump_only=True,
        metadata={'example': 2345},
    )

    text = fields.String(
        attribute='quote',
        metadata={
            'example': 'Some memorable quote',
        },
    )

    comment = fields.String(
        attribute='text',
        metadata={
            'example': 'My comment on this memorable quote',
        },
    )

    ranges = fields.Function(
        lambda data: [
            [int(r['startOffset']), int(r['endOffset'])] for r in data.get('ranges', [])
        ],
        metadata={
            'example': [[100, 180]],
        },
    )

    created_at = DateTime(
        metadata={
            'description': 'When the annotation was created',
        },
    )

    updated_at = DateTime(
        metadata={
            'description': 'When the annotation was last updated',
        },
    )


class WallabagEntrySchema(WallabagSchema):
    id = fields.Integer(
        required=True,
        dump_only=True,
        metadata={'example': 1234},
    )

    url = fields.URL(
        required=True,
        metadata={
            'description': 'Original URL',
            'example': 'https://example.com/article/some-title',
        },
    )

    preview_picture = fields.URL(
        metadata={
            'description': 'Preview picture URL',
            'example': 'https://example.com/article/some-title.jpg',
        },
    )

    is_archived = fields.Boolean()
    is_starred = fields.Boolean()
    is_public = fields.Boolean()
    mimetype = fields.String(
        metadata={
            'example': 'text/html',
        },
    )

    title = fields.String(
        metadata={
            'description': 'Title of the saved page',
        },
    )

    content = fields.String(
        metadata={
            'description': 'Parsed content',
        }
    )

    language = fields.String(
        metadata={
            'example': 'en',
        }
    )

    annotations = fields.List(fields.Nested(WallabagAnnotationSchema))

    published_by = fields.List(
        fields.String,
        metadata={
            'example': ['Author 1', 'Author 2'],
        },
    )

    tags = fields.Function(
        lambda data: [tag['label'] for tag in data.get('tags', [])],
        metadata={
            'example': ['tech', 'programming'],
        },
    )

    reading_time = fields.Integer(
        metadata={
            'description': 'Estimated reading time, in minutes',
            'example': 10,
        }
    )

    created_at = DateTime(
        metadata={
            'description': 'When the entry was created',
        },
    )

    updated_at = DateTime(
        metadata={
            'description': 'When the entry was last updated',
        },
    )

    starred_at = DateTime(
        metadata={
            'description': 'If the entry is starred, when was it last marked',
        },
    )

    published_at = DateTime(
        metadata={
            'description': 'When the entry was initially published',
        },
    )



from platypush.message.event import Event


class NgrokEvent(Event):
    """
    ``ngrok`` base event.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NgrokProcessStartedEvent(NgrokEvent):
    """
    Event triggered when the ``ngrok`` process is started.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NgrokTunnelStartedEvent(NgrokEvent):
    """
    Event triggered when a tunnel is started.
    """
    def __init__(self, *args, name: str, url: str, protocol: str, **kwargs):
        super().__init__(*args, name=name, url=url, protocol=protocol, **kwargs)


class NgrokTunnelStoppedEvent(NgrokEvent):
    """
    Event triggered when a tunnel is stopped.
    """
    def __init__(self, *args, name: str, url: str, protocol: str, **kwargs):
        super().__init__(*args, name=name, url=url, protocol=protocol, **kwargs)


class NgrokProcessStoppedEvent(NgrokEvent):
    """
    Event triggered when the ngrok process is stopped.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from abc import ABC, abstractmethod
from typing import Any

from . import EntityManager


class WriteableEntityManager(EntityManager, ABC):
    """
    Base class for integrations that support entities whose values can be set.
    """

    @abstractmethod
    def set(self, entity: str, value: Any, **kwargs):
        """
        Set the value of an entity.

        :param entity: The entity to set the value for. It's usually the ID of
            the entity provided by the plugin.
        :param value: The value to set the entity to.
        """
        raise NotImplementedError()


class SwitchEntityManager(WriteableEntityManager, ABC):
    """
    Base class for integrations that support binary switches.
    """

    @abstractmethod
    def on(self, *_, **__):
        """Turn on a device"""
        raise NotImplementedError()

    @abstractmethod
    def off(self, *_, **__):
        """Turn off a device"""
        raise NotImplementedError()

    @abstractmethod
    def toggle(self, *_, **__):
        """Toggle the state of a device (on->off or off->on)"""
        raise NotImplementedError()

    def set(self, entity: str, value: Any, **kwargs):
        method = self.on if value else self.off
        return method(entity, **kwargs)


class DimmerEntityManager(WriteableEntityManager, ABC):
    """
    Base class for integrations that support dimmers/multi-level switches.
    """


class EnumSwitchEntityManager(WriteableEntityManager, ABC):
    """
    Base class for integrations that support switches with a pre-defined,
    enum-like set of possible values.
    """



from sqlalchemy import Column, Integer, ForeignKey, Boolean, String

from platypush.common.db import is_defined

from . import Entity


if not is_defined('assistant'):

    class Assistant(Entity):
        """
        Base class for voice assistant entities.
        """

        __tablename__ = 'assistant'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )
        last_query = Column(String)
        last_response = Column(String)
        conversation_running = Column(Boolean)
        is_muted = Column(Boolean, default=False)
        is_detecting = Column(Boolean, default=True)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema


class WifiScanResultSchema(Schema):
    """
    Schema for Wi-Fi scan results.
    """

    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    essid = fields.Str(
        required=True,
        metadata={
            "description": "ESSID of the Wi-Fi network.",
            "example": "MyNetwork",
        },
    )

    bssid = fields.Str(
        required=True,
        metadata={
            "description": "BSSID of the Wi-Fi network.",
            "example": "00:11:22:33:44:55",
        },
    )

    channel = fields.Int(
        required=True,
        metadata={
            "description": "Channel of the Wi-Fi network.",
            "example": 6,
        },
    )

    rssi = fields.Int(
        required=True,
        metadata={
            "description": "RSSI of the Wi-Fi network.",
            "example": -50,
        },
    )

    auth_mode = fields.Int(
        required=True,
        metadata={
            "description": "Authentication mode of the Wi-Fi network.",
            "example": 3,
        },
    )

    hidden = fields.Bool(
        required=True,
        metadata={
            "description": "Whether the Wi-Fi network is hidden.",
            "example": False,
        },
    )


class WifiConfigSchema(Schema):
    """
    Schema for Wi-Fi configuration.
    """

    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    ip = fields.Str(
        required=True,
        metadata={
            "description": "IP address of the Wi-Fi interface.",
            "example": "192.168.1.10",
        },
    )

    netmask = fields.Str(
        required=True,
        metadata={
            "description": "Netmask of the Wi-Fi network.",
            "example": "255.255.255.0",
        },
    )

    gateway = fields.Str(
        required=True,
        metadata={
            "description": "Gateway of the Wi-Fi network.",
            "example": "192.168.1.1",
        },
    )

    dns = fields.Str(
        required=True,
        metadata={
            "description": "DNS server of the Wi-Fi network.",
            "example": "1.1.1.1",
        },
    )

    mac = fields.Str(
        required=True,
        metadata={
            "description": "MAC address of the Wi-Fi network.",
            "example": "00:11:22:33:44:55",
        },
    )

    active = fields.Bool(
        required=True,
        metadata={
            "description": "Whether the Wi-Fi network is active.",
            "example": True,
        },
    )

    essid = fields.Str(
        required=True,
        metadata={
            "description": "ESSID of the Wi-Fi network.",
            "example": "MyNetwork",
        },
    )

    channel = fields.Int(
        required=True,
        metadata={
            "description": "Channel of the Wi-Fi network.",
            "example": 6,
        },
    )

    hidden = fields.Bool(
        required=True,
        metadata={
            "description": "Whether the Wi-Fi network is hidden.",
            "example": False,
        },
    )



from typing import Optional, Type

from platypush.message.event.sound import SoundEvent

from ..._converters import RawOutputAudioFromFileConverter
from ._base import AudioPlayer


class AudioResourcePlayer(AudioPlayer):
    """
    A ``AudioResourcePlayer`` thread is responsible for playing an audio
    resource - either a file or a URL.
    """

    @property
    def _audio_converter_type(self) -> Type[RawOutputAudioFromFileConverter]:
        return RawOutputAudioFromFileConverter

    @property
    def _converter_args(self) -> dict:
        return {
            'infile': self.infile,
            'output_format': self.output_format,
            **super()._converter_args,
        }

    @property
    def _converter_stdin(self) -> Optional[int]:
        return None

    def _notify(self, event_type: Type[SoundEvent], **kwargs):
        return super()._notify(event_type, resource=self.infile, **kwargs)


# vim:sw=4:ts=4:et:



import logging
from abc import ABC, abstractmethod
from queue import Empty, Queue
from threading import Event, RLock, Thread, get_ident
from typing import Any, Optional, Sequence

from platypush.message.event.assistant import AssistantEvent

from .._context import ConversationContext


class BaseProcessor(ABC, Thread):
    """
    Base speech processor class. It is implemented by the ``SttProcessor`` and
    the ``IntentProcessor`` classes.
    """

    def __init__(
        self,
        *args,
        stop_event: Event,
        enabled: bool = True,
        conversation_timeout: Optional[float] = None,
        **kwargs,
    ):
        super().__init__(*args, name=f'picovoice:{self.__class__.__name__}', **kwargs)

        self.logger = logging.getLogger(self.name)
        self._enabled = enabled
        self._audio_queue = Queue()
        self._stop_event = stop_event
        self._ctx = ConversationContext(timeout=conversation_timeout)
        self._event_queue = Queue()
        # This event is set if the upstream processor is waiting for an event
        # from this processor
        self._event_wait = Event()
        # This event is set when the processor is done with the audio
        # processing and it's ready to accept a new audio frame
        self._processing_done = Event()
        self._processing_done.set()
        self._state_lock = RLock()

    def should_stop(self) -> bool:
        return self._stop_event.is_set()

    def wait_stop(self, timeout: Optional[float] = None) -> bool:
        return self._stop_event.wait(timeout)

    def enqueue(self, audio: Sequence[int]):
        if not self._enabled:
            return

        self._event_wait.set()
        self._processing_done.clear()
        self._audio_queue.put_nowait(audio)

    def reset(self) -> Optional[Any]:
        """
        Reset any pending context.
        """
        if not self._enabled:
            return

        with self._state_lock:
            self._ctx.reset()
            self._event_queue.queue.clear()
            self._event_wait.clear()
            self._processing_done.set()

    @property
    def processing_done(self) -> Event:
        return self._processing_done

    @property
    @abstractmethod
    def _model_path(self) -> Optional[str]:
        """
        Return the model path.
        """

    @property
    @abstractmethod
    def sample_rate(self) -> int:
        """
        :return: The sample rate wanted by Cheetah/Rhino.
        """

    @property
    @abstractmethod
    def frame_length(self) -> int:
        """
        :return: The frame length wanted by Cheetah/Rhino.
        """

    def last_event(self) -> Optional[AssistantEvent]:
        """
        :return: The latest event that was processed by the processor.
        """
        with self._state_lock:
            evt = None
            try:
                while True:
                    evt = self._event_queue.get_nowait()
            except Empty:
                pass

            if evt:
                self._event_wait.clear()

            return evt

    @abstractmethod
    def process(self, audio: Sequence[int]) -> Optional[AssistantEvent]:
        """
        Process speech events from a raw audio input.
        """

    def run(self):
        super().run()
        if not self._enabled:
            self.wait_stop()

        self.reset()
        self._processing_done.clear()
        self.logger.info('Processor started: %s', self.name)

        while not self.should_stop():
            audio = self._audio_queue.get()

            # The thread is stopped when it receives a None object
            if audio is None:
                break

            # Don't process the audio if the upstream processor is not waiting
            # for an event
            if not self._event_wait.is_set():
                continue

            try:
                self._processing_done.clear()
                event = self.process(audio)

                if event:
                    self.logger.debug(
                        'Dispatching event processed from %s: %s', self.name, event
                    )
                    self._event_queue.put_nowait(event)
                    self._processing_done.set()
            except Exception as e:
                self.logger.error(
                    'An error occurred while processing the audio on %s: %s',
                    self.name,
                    e,
                    exc_info=e,
                )
                self.wait_stop(timeout=1)
                self._processing_done.set()
                continue

        self._processing_done.set()
        self.reset()
        self.logger.info('Processor stopped: %s', self.name)

    def stop(self):
        if not self._enabled:
            return

        self._audio_queue.put_nowait(None)
        if self.is_alive() and self.ident != get_ident():
            self.logger.debug('Stopping %s', self.name)
            self.join()

    def on_conversation_start(self):
        self._ctx.start()

    def on_conversation_end(self):
        self.reset()

    def on_conversation_reset(self):
        self.reset()



from abc import ABC, abstractmethod
import hashlib
import logging
import os
from typing import Generator, Optional

from platypush.message import JSONAble


class MediaHandler(JSONAble, ABC):
    """
    Abstract class to manage media handlers that can be streamed over the HTTP
    server through the `/media` endpoint.
    """

    prefix_handlers = []

    def __init__(
        self,
        source: str,
        *_,
        filename: Optional[str] = None,
        mime_type: str = 'application/octet-stream',
        name: Optional[str] = None,
        url: Optional[str] = None,
        subtitles: Optional[str] = None,
        **__,
    ):
        matched_handlers = [
            hndl for hndl in self.prefix_handlers if source.startswith(hndl)
        ]

        if not matched_handlers:
            raise AttributeError(
                f'No matched handlers found for source "{source}" '
                f'through {self.__class__.__name__}. Supported handlers: {self.prefix_handlers}'
            )

        self.name = name
        self.path = None
        self.filename = filename
        self.source = source
        self.url = url
        self.mime_type = mime_type
        self.subtitles = subtitles
        self.content_length = 0
        self._matched_handler = matched_handlers[0]

    @classmethod
    def build(cls, source: str, *args, **kwargs) -> 'MediaHandler':
        errors = {}

        for hndl_class in supported_handlers:
            try:
                return hndl_class(source, *args, **kwargs)
            except Exception as e:
                logging.exception(e)
                errors[hndl_class.__name__] = str(e)

        if os.path.exists(source):
            source = f'file://{source}'

        raise AttributeError(
            f'The source {source} has no handlers associated. Errors: {errors}'
        )

    @abstractmethod
    def get_data(
        self,
        from_bytes: Optional[int] = None,
        to_bytes: Optional[int] = None,
        chunk_size: Optional[int] = None,
    ) -> Generator[bytes, None, None]:
        raise NotImplementedError()

    @property
    def media_id(self) -> str:
        """
        :returns: The unique ID of the media handler.
        """
        return self.get_media_id(self.source)

    def set_subtitles(self, subtitles_file: Optional[str]):
        self.subtitles = subtitles_file

    def remove_subtitles(self):
        self.subtitles = None

    def __iter__(self):
        """
        Iterate over the attributes of the media handler.
        """
        for attr in [
            'name',
            'source',
            'mime_type',
            'url',
            'subtitles',
            'prefix_handlers',
            'media_id',
        ]:
            if hasattr(self, attr):
                yield attr, getattr(self, attr)

    @staticmethod
    def get_media_id(source: str) -> str:
        """
        :returns: The ID of a media file given its source.
        """
        return hashlib.sha1(source.encode()).hexdigest()

    def to_json(self):
        """
        :returns: A dictionary representation of the media handler.
        """
        return dict(self)


from .file import FileHandler  # noqa

__all__ = ['MediaHandler', 'FileHandler']


supported_handlers = [eval(hndl) for hndl in __all__ if hndl != MediaHandler.__name__]


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional, Union


class BridgeState(Enum):
    """
    Known bridge states.
    """

    ONLINE = 'online'
    OFFLINE = 'offline'


@dataclass
class ZigbeeDevices:
    """
    Cached information about the devices on the network.
    """

    by_address: Dict[str, dict] = field(default_factory=dict)
    by_name: Dict[str, dict] = field(default_factory=dict)

    def __contains__(self, name: str) -> bool:
        """
        :return: True if the device with the given name exists in the cache.
        """
        return name in self.by_name or name in self.by_address

    def get(self, name: str) -> Optional[dict]:
        """
        Retrieves a cached device record either by name or by address.
        """
        return self.by_address.get(name, self.by_name.get(name))

    def add(self, device: dict):
        """
        Adds a device record to the cache.
        """
        if device.get('ieee_address'):
            self.by_address[device['ieee_address']] = device
        if device.get('friendly_name'):
            self.by_name[device['friendly_name']] = device
        if not device.get('state'):
            device['state'] = {}

    def remove(self, device: Union[str, dict]):
        """
        Removes a device record from the cache.
        """
        if isinstance(device, str):
            dev = self.get(device)
            if not dev:
                return  # No such device
        else:
            dev = device

        if dev.get('ieee_address'):
            self.by_address.pop(dev['ieee_address'], None)

        if dev.get('friendly_name'):
            self.by_name.pop(dev['friendly_name'], None)

    def set_state(self, device: str, state: dict):
        """
        Updates the state of a device in the cache.

        :param device: Name or address of the device.
        :param state: Map containing the new state.
        """
        dev = self.get(device)
        if not dev:
            return

        dev['state'] = state


@dataclass
class ZigbeeState:
    """
    Cached information about the devices and groups on the network.
    """

    devices: ZigbeeDevices = field(default_factory=ZigbeeDevices)
    groups: Dict[str, dict] = field(default_factory=dict)
    bridge_state: BridgeState = BridgeState.OFFLINE



from logging import getLogger
import os
import stat
from threading import Event
from time import time
from typing import Iterable, List, Optional, Union

from .._model import AudioDevice, DeviceType, StreamType
from .._streams import AudioPlayer, AudioRecorder, AudioThread
from ._device import DeviceManager
from ._stream import StreamManager


class AudioManager:
    """
    The audio manager is responsible for managing multiple audio controllers and
    their access to audio resources.

    It main purpose is to act as a proxy/facade between the high-level audio
    plugin and the audio functionalities (allocating streams, managing the state
    of the player and recorder processes, etc.).
    """

    _default_signal_timeout = 2

    def __init__(
        self,
        should_stop: Event,
        input_blocksize: int,
        output_blocksize: int,
        input_device: Optional[DeviceType] = None,
        output_device: Optional[DeviceType] = None,
        queue_size: Optional[int] = None,
    ):
        """
        :param should_stop: Event to synchronize the audio manager stop.
        :param input_blocksize: Block size for the input stream.
        :param output_blocksize: Block size for the output stream.
        :param input_device: Default device to use for the input stream.
        :param output_device: Default device to use for the output stream.
        :param queue_size: Maximum size of the audio queues.
        """
        self._should_stop = should_stop
        self._device_manager = DeviceManager(
            input_device=input_device, output_device=output_device
        )

        self._stream_manager = StreamManager(device_manager=self._device_manager)
        self.logger = getLogger(__name__)
        self.input_blocksize = input_blocksize
        self.output_blocksize = output_blocksize
        self.queue_size = queue_size

    def create_player(
        self,
        device: DeviceType,
        channels: int,
        volume: float,
        infile: Optional[str] = None,
        sound: Optional[Union[dict, Iterable[dict]]] = None,
        duration: Optional[float] = None,
        sample_rate: Optional[int] = None,
        dtype: str = 'int16',
        format: Optional[str] = None,  # pylint: disable=redefined-builtin
        blocksize: Optional[int] = None,
        latency: Union[float, str] = 'high',
        stream_name: Optional[str] = None,
    ) -> AudioPlayer:
        """
        Create an audio player thread.

        :param device: Audio device to use.
        :param channels: Number of output channels.
        :param volume: Output volume, between 0 and 100.
        :param infile: File or URL to play.
        :param sound: Alternatively to a file/URL, you can play synthetic
            sounds.
        :param duration: Duration of the stream in seconds.
        :param sample_rate: Sample rate of the stream.
        :param dtype: Data type of the stream.
        :param format: Output format of the stream.
        :param blocksize: Block size of the stream.
        :param latency: Latency of the stream.
        :param stream_name: Name of the stream.
        """
        dev = self._device_manager.get_device(device, type=StreamType.OUTPUT)
        player = AudioPlayer.build(
            device=device,
            infile=infile,
            sound=sound,
            duration=duration,
            volume=volume,
            sample_rate=sample_rate or dev.default_samplerate,
            dtype=dtype,
            blocksize=blocksize or self.output_blocksize,
            latency=latency,
            channels=channels,
            output_format=format,
            queue_size=self.queue_size,
            should_stop=self._should_stop,
        )

        self._stream_manager.register(
            player, dev, StreamType.OUTPUT, stream_name=stream_name
        )
        return player

    def create_recorder(
        self,
        device: DeviceType,
        output_device: Optional[DeviceType] = None,
        fifo: Optional[str] = None,
        outfile: Optional[str] = None,
        duration: Optional[float] = None,
        sample_rate: Optional[int] = None,
        dtype: str = 'int16',
        blocksize: Optional[int] = None,
        latency: Union[float, str] = 'high',
        channels: int = 1,
        volume: float = 100,
        redis_queue: Optional[str] = None,
        format: str = 'wav',  # pylint: disable=redefined-builtin
        stream: bool = True,
        stream_name: Optional[str] = None,
        play_audio: bool = False,
    ) -> AudioRecorder:
        """
        Create an audio recorder thread.

        :param device: Audio device to use.
        :param output_device: Output device to use.
        :param fifo: Path to an output FIFO file to use to synchronize the audio
            to other processes.
        :param outfile: Optional output file for the recorded audio.
        :param duration: Duration of the recording in seconds.
        :param sample_rate: Sample rate of the stream.
        :param dtype: Data type of the stream.
        :param blocksize: Block size of the stream.
        :param latency: Latency of the stream.
        :param channels: Number of output channels.
        :param volume: Input volume, between 0 and 100.
        :param redis_queue: Name of the Redis queue to use.
        :param format: Format of the recorded audio.
        :param stream: Whether to stream the recorded audio.
        :param play_audio: Whether to play the recorded audio in real-time.
        :param stream_name: Name of the stream.
        """
        blocksize = blocksize or self.input_blocksize
        dev = self._device_manager.get_device(device, type=StreamType.OUTPUT)

        if fifo:
            fifo = os.path.expanduser(fifo)
            if os.path.exists(fifo) and stat.S_ISFIFO(os.stat(fifo).st_mode):
                self.logger.info('Removing previous input stream FIFO %s', fifo)
                os.unlink(fifo)

            os.mkfifo(fifo, 0o644)
            outfile = fifo
        elif outfile:
            outfile = os.path.expanduser(outfile)

        outfile = outfile or fifo or os.devnull
        recorder = AudioRecorder(
            device=(
                (
                    dev.index,
                    self._device_manager.get_device(
                        type=StreamType.OUTPUT, device=output_device
                    ).index,
                )
                if play_audio
                else dev.index
            ),
            outfile=outfile,
            duration=duration,
            sample_rate=sample_rate or dev.default_samplerate,
            dtype=dtype,
            blocksize=blocksize,
            latency=latency,
            output_format=format,
            channels=channels,
            volume=volume,
            redis_queue=redis_queue,
            stream=stream,
            audio_pass_through=play_audio,
            queue_size=self.queue_size,
            should_stop=self._should_stop,
        )

        self._stream_manager.register(
            recorder, dev, StreamType.INPUT, stream_name=stream_name
        )
        return recorder

    def get_device(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
    ) -> AudioDevice:
        """
        Proxy to ``self._device_manager.get_device``.
        """
        return self._device_manager.get_device(device=device, type=type)

    def get_devices(
        self,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
    ) -> List[AudioDevice]:
        """
        Proxy to ``self._device_manager.get_devices``.
        """
        return self._device_manager.get_devices(type=type)

    def get_streams(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
        streams: Optional[Iterable[Union[str, int]]] = None,
    ) -> List[AudioThread]:
        """
        Proxy to ``self._stream_manager.get``.
        """
        return self._stream_manager.get(device=device, type=type, streams=streams)

    def stop_audio(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
        streams: Optional[Iterable[Union[str, int]]] = None,
        timeout: Optional[float] = 2,
    ):
        """
        Stops audio sessions.

        :param device: Filter by host audio device.
        :param type: Filter by stream type (input or output).
        :param streams: Filter by stream indices/names.
        :param timeout: Wait timeout in seconds.
        """
        streams_to_stop = self._stream_manager.get(device, type, streams=streams)

        # Send the stop signals
        for audio_thread in streams_to_stop:
            audio_thread.notify_stop()

        # Wait for termination (with timeout)
        wait_start = time()
        for audio_thread in streams_to_stop:
            audio_thread.join(
                timeout=(
                    max(0, timeout - (time() - wait_start))
                    if timeout is not None
                    else None
                )
            )

        # Remove references
        for audio_thread in streams_to_stop:
            self._stream_manager.unregister(audio_thread)

    def pause_audio(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
        streams: Optional[Iterable[Union[str, int]]] = None,
    ):
        """
        Pauses/resumes audio sessions.

        :param device: Filter by host audio device.
        :param type: Filter by stream type (input or output).
        :param streams: Filter by stream indices/names.
        """
        streams_to_pause = self._stream_manager.get(device, type, streams=streams)

        # Send the pause toggle signals
        for audio_thread in streams_to_pause:
            audio_thread.notify_pause()

    def set_volume(
        self,
        volume: float,
        device: Optional[DeviceType] = None,
        streams: Optional[Iterable[Union[str, int]]] = None,
    ):
        """
        :param volume: New volume, between 0 and 100.
        :param device: Set the volume only on the specified device (default:
            all).
        :param streams: Set the volume only on the specified list of stream
            indices/names (default: all).
        """
        stream_objs = self._stream_manager.get(device=device, streams=streams)

        for stream in stream_objs:
            stream.volume = volume



from ._base import TorrentSearchProvider
from ._popcorntime import PopcornTimeSearchProvider
from ._torrents_csv import TorrentsCsvSearchProvider


__all__ = [
    'TorrentSearchProvider',
    'PopcornTimeSearchProvider',
    'TorrentsCsvSearchProvider',
]


# vim:sw=4:ts=4:et:



from enum import Enum


class Protocol(Enum):
    """
    Models a Bluetooth protocol.
    """

    UNKNOWN = 'UNKNOWN'
    RFCOMM = 'RFCOMM'
    L2CAP = 'L2CAP'
    TCP = 'TCP'
    UDP = 'UDP'
    SDP = 'SDP'
    BNEP = 'BNEP'
    TCS_BIN = 'TCS-BIN'
    TCS_AT = 'TCS-AT'
    OBEX = 'OBEX'
    IP = 'IP'
    FTP = 'FTP'
    HTTP = 'HTTP'
    WSP = 'WSP'
    UPNP = 'UPNP'
    HIDP = 'HIDP'
    AVCTP = 'AVCTP'
    AVDTP = 'AVDTP'
    CMTP = 'CMTP'
    UDI_C_PLANE = 'UDI_C-Plane'
    HardCopyControlChannel = 'HardCopyControlChannel'
    HardCopyDataChannel = 'HardCopyDataChannel'
    HardCopyNotification = 'HardCopyNotification'

    def __str__(self) -> str:
        """
        Only returns the value of the enum.
        """
        return self.value

    def __repr__(self) -> str:
        """
        Only returns the value of the enum.
        """
        return str(self)



from marshmallow_dataclass import class_schema

from ._base import TemperatureBaseSchema
from ._model import Temperature


TemperatureSchema = class_schema(Temperature, base_schema=TemperatureBaseSchema)



import base64
import os
import re
from queue import Empty, Queue
import threading
import time
from typing import (
    Any,
    Collection,
    Dict,
    Final,
    List,
    Optional,
    Union,
    Type,
)

from platypush.common import StoppableThread
from platypush.context import get_bus, get_plugin
from platypush.entities import (
    EnumSwitchEntityManager,
    get_entities_engine,
)
from platypush.entities.bluetooth import BluetoothDevice, BluetoothService
from platypush.message.event.bluetooth import (
    BluetoothScanPausedEvent,
    BluetoothScanResumedEvent,
)
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.db import DbPlugin

from ._ble import BLEManager
from ._cache import EntityCache
from ._legacy import LegacyManager
from ._types import DevicesBlacklist, RawServiceClass
from ._manager import BaseBluetoothManager


# pylint: disable=too-many-ancestors
class BluetoothPlugin(RunnablePlugin, EnumSwitchEntityManager):
    """
    Plugin to interact with Bluetooth devices.

    This plugin uses `Bleak <https://github.com/hbldh/bleak>`_ to interact
    with the Bluetooth stack and `Theengs <https://github.com/theengs/decoder>`_
    to map the services exposed by the devices into native entities.

    The full list of devices natively supported can be found
    `here <https://decoder.theengs.io/devices/devices_by_brand.html>`_.

    It also supports legacy Bluetooth services, as well as the transfer of
    files.

    Note that the support for Bluetooth low-energy devices requires a Bluetooth
    adapter compatible with the Bluetooth 5.0 specification or higher.
    """

    _default_connect_timeout: Final[int] = 20
    """ Default connection timeout (in seconds) """

    _default_scan_duration: Final[float] = 10.0
    """ Default duration of a discovery session (in seconds) """

    def __init__(
        self,
        interface: Optional[str] = None,
        connect_timeout: float = _default_connect_timeout,
        service_uuids: Optional[Collection[RawServiceClass]] = None,
        scan_paused_on_start: bool = False,
        poll_interval: float = _default_scan_duration,
        exclude_known_noisy_beacons: bool = True,
        ignored_device_addresses: Optional[Collection[str]] = None,
        ignored_device_names: Optional[Collection[str]] = None,
        ignored_device_manufacturers: Optional[Collection[str]] = None,
        **kwargs,
    ):
        """
        :param interface: Name of the Bluetooth interface to use (e.g. ``hci0``
            on Linux). Default: first available interface.
        :param connect_timeout: Timeout in seconds for the connection to a
            Bluetooth device. Default: 20 seconds.
        :param service_uuids: List of service UUIDs to discover.
            Default: all.
        :param scan_paused_on_start: If ``True``, the plugin will not the
            scanning thread until :meth:`.scan_resume` is called (default:
            ``False``).
        :param exclude_known_noisy_beacons: Exclude BLE beacons from devices
            known for being very noisy. It mainly includes tracking services on
            Google, Apple, Microsoft and Samsung devices. These devices are
            also known for refreshing their MAC address very frequently, which
            may result in a large (and constantly increasing) list of devices.
            Disable this flag if you need to track BLE beacons from these
            devices, but beware that you may need periodically clean up your
            list of scanned devices.
        :param ignored_device_addresses: List of device addresses to ignore.
        :param ignored_device_names: List of device names to ignore.
        :param ignored_device_manufacturers: List of device manufacturers to
            ignore.
        """
        kwargs['poll_interval'] = poll_interval
        super().__init__(**kwargs)

        self._interface: Optional[str] = interface
        """ Default Bluetooth interface to use """
        self._connect_timeout: float = connect_timeout
        """ Connection timeout in seconds """
        self._service_uuids: Collection[RawServiceClass] = service_uuids or []
        """ UUIDs to discover """
        self._scan_lock = threading.RLock()
        """ Lock to synchronize scanning access to the Bluetooth device """
        self._scan_enabled = threading.Event()
        """ Event used to enable/disable scanning """
        self._device_queue: Queue[BluetoothDevice] = Queue()
        """
        Queue used by the Bluetooth managers to published the discovered
        Bluetooth devices.
        """
        self._device_cache = EntityCache()
        """
        Cache of the devices discovered by the plugin.
        """
        self._excluded_known_noisy_beacons = exclude_known_noisy_beacons
        """ Exclude known noisy BLE beacons. """

        self._blacklist = DevicesBlacklist(
            addresses=set(ignored_device_addresses or []),
            names=set(ignored_device_names or []),
            manufacturers=set(ignored_device_manufacturers or []),
        )
        """ Blacklist rules for the devices to ignore. """

        self._managers: Dict[Type[BaseBluetoothManager], BaseBluetoothManager] = {}
        """
        Bluetooth managers threads, one for BLE devices and one for non-BLE
        devices.
        """

        self._scan_controller_timer: Optional[threading.Timer] = None
        """ Timer used to temporarily pause the discovery process """

        if not scan_paused_on_start:
            self._scan_enabled.set()

    def _refresh_cache(self) -> None:
        # Wait for the entities engine to start
        get_entities_engine().wait_start()

        with get_plugin(DbPlugin).get_session(
            autoflush=False, autocommit=False, expire_on_commit=False
        ) as session:
            existing_devices = [d.copy() for d in session.query(BluetoothDevice).all()]

        for dev in existing_devices:
            self._device_cache.add(dev)

    def _init_bluetooth_managers(self):
        """
        Initializes the Bluetooth managers threads.
        """
        manager_args = {
            'interface': self._interface,
            'poll_interval': self.poll_interval,
            'connect_timeout': self._connect_timeout,
            'stop_event': self._should_stop,
            'scan_lock': self._scan_lock,
            'scan_enabled': self._scan_enabled,
            'device_queue': self._device_queue,
            'service_uuids': list(map(BluetoothService.to_uuid, self._service_uuids)),
            'device_cache': self._device_cache,
            'exclude_known_noisy_beacons': self._excluded_known_noisy_beacons,
            'blacklist': self._blacklist,
        }

        self._managers = {
            BLEManager: BLEManager(**manager_args),
            LegacyManager: LegacyManager(**manager_args),
        }

    def _scan_state_set(self, state: bool, duration: Optional[float] = None):
        """
        Set the state of the scanning process.

        :param state: ``True`` to enable the scanning process, ``False`` to
            disable it.
        :param duration: The duration of the pause (in seconds) or ``None``.
        """

        def timer_callback():
            if state:
                self.scan_pause()
            else:
                self.scan_resume()

            self._scan_controller_timer = None

        with self._scan_lock:
            if not state and self._scan_enabled.is_set():
                get_bus().post(BluetoothScanPausedEvent(duration=duration))
            elif state and not self._scan_enabled.is_set():
                get_bus().post(BluetoothScanResumedEvent(duration=duration))

            if state:
                self._scan_enabled.set()
            else:
                self._scan_enabled.clear()

            if duration and not self._scan_controller_timer:
                self._scan_controller_timer = threading.Timer(duration, timer_callback)
                self._scan_controller_timer.start()

    def _cancel_scan_controller_timer(self):
        """
        Cancels a scan controller timer if scheduled.
        """
        if self._scan_controller_timer:
            self._scan_controller_timer.cancel()

    def _manager_by_device(
        self,
        device: BluetoothDevice,
        port: Optional[int] = None,
        service_uuid: Optional[Union[str, RawServiceClass]] = None,
    ) -> BaseBluetoothManager:
        """
        :param device: A discovered Bluetooth device.
        :param port: The port to connect to.
        :param service_uuid: The UUID of the service to connect to.
        :return: The manager associated with the device (BLE or legacy).
        """
        # No port nor service UUID -> use the BLE manager for direct connection
        if not (port or service_uuid):
            return self._managers[BLEManager]

        uuid = BluetoothService.to_uuid(service_uuid) if service_uuid else None
        matching_services = (
            [srv for srv in device.services if srv.port == port]
            if port
            else [srv for srv in device.services if srv.uuid == uuid]
        )

        if not matching_services:
            # It could be a GATT characteristic, so try BLE
            return self._managers[BLEManager]

        srv = matching_services[0]
        return (
            self._managers[BLEManager] if srv.is_ble else self._managers[LegacyManager]
        )

    def _get_device(self, device: str, _fail_if_not_cached=False) -> BluetoothDevice:
        """
        Get a device by its address or name, and scan for it if it's not
        cached.
        """
        # If device is a compound entity ID in the format
        # ``<mac_address>:<service>``, then split the MAC address part
        m = re.match(r'^(([0-9a-f]{2}:){6}):.*', device, re.IGNORECASE)
        if m:
            device = m.group(1).rstrip(':')

        dev = self._device_cache.get(device)
        if dev:
            return dev

        assert not _fail_if_not_cached, f'Device {device} not found'
        self.logger.info('Scanning for unknown device %s', device)
        self.scan()
        return self._get_device(device, _fail_if_not_cached=True)

    @action
    def connect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[Union[RawServiceClass, str]] = None,
        interface: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        """
        Pair and connect to a device by address or name.

        :param device: The device address or name.
        :param port: The port to connect to. Either ``port`` or
            ``service_uuid`` is required for non-BLE devices.
        :param service_uuid: The UUID of the service to connect to. Either
            ``port`` or ``service_uuid`` is required for non-BLE devices.
        :param interface: The Bluetooth interface to use (it overrides the
            default ``interface``).
        :param timeout: The connection timeout in seconds (it overrides the
            default ``connect_timeout``).
        """
        dev = self._get_device(device)
        manager = self._manager_by_device(dev, port=port, service_uuid=service_uuid)
        uuid = BluetoothService.to_uuid(service_uuid) if service_uuid else None
        manager.connect(
            dev.address,
            port=port,
            service_uuid=uuid,
            interface=interface,
            timeout=timeout,
        )

    @action
    def disconnect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
    ):
        """
        Close an active connection to a device.

        Note that this method can only close connections that have been
        initiated by the application. It can't close connections owned by
        other applications or agents.

        :param device: The device address or name.
        :param port: If connected to a non-BLE device, the optional port to
            disconnect.
        :param service_uuid: The optional UUID of the service to disconnect
            from, for non-BLE devices.
        """
        dev = self._get_device(device)
        uuid = BluetoothService.to_uuid(service_uuid) if service_uuid else None
        err = None
        success = False

        for manager in self._managers.values():
            try:
                manager.disconnect(dev.address, port=port, service_uuid=uuid)
                success = True
            except Exception as e:
                err = e

        assert success, f'Could not disconnect from {device}: {err}'

    @action
    def scan_pause(self, duration: Optional[float] = None):
        """
        Pause the scanning thread.

        :param duration: For how long the scanning thread should be paused
            (default: null = indefinitely).
        """
        self._scan_state_set(False, duration)

    @action
    def scan_resume(self, duration: Optional[float] = None):
        """
        Resume the scanning thread, if inactive.

        :param duration: For how long the scanning thread should be running
            (default: null = indefinitely).
        """
        self._scan_state_set(True, duration)

    @action
    def scan(
        self,
        duration: Optional[float] = None,
        devices: Optional[Collection[str]] = None,
        service_uuids: Optional[Collection[RawServiceClass]] = None,
    ) -> List[BluetoothDevice]:
        """
        Scan for Bluetooth devices nearby and return the results as a list of
        entities.

        :param duration: Scan duration in seconds (default: same as the plugin's
            `poll_interval` configuration parameter)
        :param devices: List of device addresses or names to scan for.
        :param service_uuids: List of service UUIDs to discover. Default: all.
        """
        scanned_device_addresses = set()
        duration = duration or self.poll_interval or self._default_scan_duration
        uuids = {BluetoothService.to_uuid(uuid) for uuid in (service_uuids or [])}

        for manager in self._managers.values():
            scanned_device_addresses.update(
                [
                    device.address
                    for device in manager.scan(duration=duration // len(self._managers))
                    if (not uuids or any(srv.uuid in uuids for srv in device.services))
                    and (
                        not devices
                        or device.address in devices
                        or device.name in devices
                    )
                ]
            )

        with get_plugin(DbPlugin).get_session(
            autoflush=False, autocommit=False, expire_on_commit=False
        ) as session:
            return [
                d.copy()
                for d in session.query(BluetoothDevice).all()
                if d.address in scanned_device_addresses
            ]

    @action
    def read(
        self,
        device: str,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ) -> str:
        """
        Read a message from a device.

        :param device: Name or address of the device to read from.
        :param service_uuid: Service UUID.
        :param interface: Bluetooth adapter name to use (default configured if None).
        :param connect_timeout: Connection timeout in seconds (default: same as the
            configured `connect_timeout`).
        :return: The base64-encoded response received from the device.
        """
        dev = self._get_device(device)
        uuid = BluetoothService.to_uuid(service_uuid)
        manager = self._manager_by_device(dev, service_uuid=uuid)
        data = manager.read(
            dev.address, uuid, interface=interface, connect_timeout=connect_timeout
        )
        return base64.b64encode(data).decode()

    @action
    def write(
        self,
        device: str,
        data: str,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ):
        """
        Writes data to a device

        :param device: Name or address of the device to read from.
        :param data: Data to be written, as a base64-encoded string.
        :param service_uuid: Service UUID.
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param connect_timeout: Connection timeout in seconds (default: same as the
            configured `connect_timeout`).
        """
        binary_data = base64.b64decode(data.encode())
        dev = self._get_device(device)
        uuid = BluetoothService.to_uuid(service_uuid)
        manager = self._manager_by_device(dev, service_uuid=uuid)
        manager.write(
            dev.address,
            binary_data,
            service_uuid=uuid,
            interface=interface,
            connect_timeout=connect_timeout,
        )

    @action
    def set(self, entity: str, value: Any, **_):
        """
        Set the value of an entity.

        This is currently only supported for SwitchBot devices, where the value
        can be one among ``on``, ``off`` and ``press``.

        :param entity: The entity to set the value for. It can be the full
            entity ID in the format ``<mac-address>::<service>``, or just
            the MAC address if the plugin supports it.
        :param value: The value to set the entity to.
        """
        device = self._get_device(entity)
        matching_plugin = next(
            iter(
                plugin
                for manager in self._managers.values()
                for plugin in manager.plugins
                if plugin.supports_device(device)
            ),
            None,
        )

        assert (
            matching_plugin is not None
        ), f'Action `set` not supported on device {entity}'

        method = getattr(matching_plugin, 'set', None)
        assert method, f'The plugin {matching_plugin} does not support `set`'
        return method(device, value)

    @action
    def send_file(
        self,
        file: str,
        device: str,
        data: Optional[Union[str, bytes, bytearray]] = None,
        binary: bool = False,
    ):
        """
        Send a file to a device that exposes an OBEX Object Push service.

        :param file: Path of the file to be sent. If ``data`` is specified
            then ``file`` should include the proposed file on the
            receiving host.
        :param data: Alternatively to a file on disk you can send raw (string
            or binary) content.
        :param device: Device address or name.
        :param binary: Set to true if data is a base64-encoded binary string.
        """
        from ._file import FileSender

        if not data:
            file = os.path.abspath(os.path.expanduser(file))
            with open(file, 'rb') as f:
                binary_data = f.read()
        elif binary:
            binary_data = base64.b64decode(
                data.encode() if isinstance(data, str) else data
            )
        elif isinstance(data, str):
            binary_data = data.encode()
        else:
            binary_data = data

        sender = FileSender(self._managers[LegacyManager])  # type: ignore
        sender.send_file(file, device, binary_data)

    @action
    def status(
        self,
        *_,
        duration: Optional[float] = None,
        devices: Optional[Collection[str]] = None,
        service_uuids: Optional[Collection[RawServiceClass]] = None,
        **__,
    ) -> List[BluetoothDevice]:
        """
        Retrieve the status of all the devices, or the matching
        devices/services.

        If scanning is currently disabled, it will enable it and perform a
        scan.

        The differences between this method and :meth:`.scan` are:

            1. :meth:`.status` will return the status of all the devices known
               to the application, while :meth:`.scan` will return the status
               only of the devices discovered in the provided time window.

            2. :meth:`.status` will not initiate a new scan if scanning is
               already enabled (it will only return the status of the known
               devices), while :meth:`.scan` will initiate a new scan.

        :param duration: Scan duration in seconds, if scanning is disabled
            (default: same as the plugin's `poll_interval` configuration
            parameter)
        :param devices: List of device addresses or names to filter for.
            Default: all.
        :param service_uuids: List of service UUIDs to filter for. Default:
            all.
        """
        if not self._scan_enabled.is_set():
            self.scan(
                duration=duration,
                devices=devices,
                service_uuids=service_uuids,
            )

        with get_plugin(DbPlugin).get_session(
            autoflush=False, autocommit=False, expire_on_commit=False
        ) as session:
            known_devices = [
                d.copy()
                for d in session.query(BluetoothDevice).all()
                if (not devices or d.address in devices or d.name in devices)
                and (
                    not service_uuids
                    or any(str(srv.uuid) in service_uuids for srv in d.services)
                )
            ]

        # Send entity update events to keep any asynchronous clients in sync
        get_entities_engine().notify(*known_devices)
        return known_devices

    def transform_entities(
        self, entities: Collection[BluetoothDevice]
    ) -> Collection[BluetoothDevice]:
        return super().transform_entities(entities)

    def main(self):
        self._refresh_cache()
        self._init_bluetooth_managers()

        for manager in self._managers.values():
            manager.start()

        try:
            while not self.should_stop():
                try:
                    device = self._device_queue.get(timeout=1)
                except Empty:
                    continue

                device = self._device_cache.add(device)
                self.publish_entities([device], callback=self._device_cache.add)
        finally:
            self.stop()

    def stop(self):
        """
        Upon stop request, it stops any pending scans and closes all active
        connections.
        """
        super().stop()

        self._cancel_scan_controller_timer()
        self._stop_threads(self._managers.values())

    def _stop_threads(self, threads: Collection[StoppableThread], timeout: float = 5):
        """
        Set the stop events on active threads and wait for them to stop.
        """
        # Set the stop events and call `.stop`
        for thread in threads:
            if thread and thread.is_alive():
                self.logger.info('Waiting for %s to stop', thread.name)
                try:
                    thread.stop()
                except Exception as e:
                    self.logger.exception('Error while stopping %s: %s', thread.name, e)

        # Wait for the manager threads to stop
        wait_start = time.time()

        for thread in threads:
            if (
                thread
                and thread.ident != threading.current_thread().ident
                and thread.is_alive()
            ):
                thread.join(timeout=max(0, int(timeout - (time.time() - wait_start))))

                if thread and thread.is_alive():
                    self.logger.warning(
                        'Timeout while waiting for %s to stop', thread.name
                    )


__all__ = ["BluetoothPlugin"]

# vim:sw=4:ts=4:et:



from sqlalchemy import Boolean, Column, Float, ForeignKey, Integer, JSON, String

from platypush.common.db import is_defined

from . import Entity


if not is_defined('alarm'):

    class Alarm(Entity):
        """
        Alarm entity model.
        """

        __tablename__ = 'alarm'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        when = Column(String, nullable=False)
        next_run = Column(Float, nullable=True)
        enabled = Column(Boolean, nullable=False, default=True)
        state = Column(String, nullable=False, default='UNKNOWN')
        media = Column(String, nullable=True)
        media_plugin = Column(String, nullable=True)
        media_repeat = Column(Boolean, nullable=False, default=True)
        audio_volume = Column(Integer, nullable=True)
        snooze_interval = Column(Integer, nullable=True)
        dismiss_interval = Column(Integer, nullable=True)
        actions = Column(JSON, nullable=True)
        static = Column(Boolean, nullable=False, default=False)
        condition_type = Column(String, nullable=False)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



import os
from dataclasses import dataclass, field
from typing import Any, Iterable, Optional
from uuid import UUID, uuid4


class Resource:
    """
    Class to handle resources (files) to be sent through the Telegram API.
    """

    def __init__(
        self,
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
    ):
        assert file_id or url or path, 'You need to specify either file_id, url or path'
        self.file_id = file_id
        self.url = url
        self.path = path
        self.fd = None

    def __enter__(self):
        """
        Context manager to open the file and return the file descriptor.
        """
        if self.path:
            self.fd = open(os.path.abspath(os.path.expanduser(self.path)), 'rb')  # noqa
            return self.fd

        return self.file_id or self.url

    def __exit__(self, *_, **__):
        """
        If the file was opened, close it.
        """
        if self.fd:
            self.fd.close()
            self.fd = None


@dataclass
class Command:
    """
    Dataclass to represent a command to be executed on the Telegram service.
    """

    cmd: str
    args: Iterable = field(default_factory=list)
    kwargs: dict = field(default_factory=dict)
    timeout: Optional[float] = None
    id: UUID = field(default_factory=uuid4)

    def is_end_of_service(self) -> bool:
        """
        Check if a command is the end-of-service command.
        """
        return is_end_of_service(self)


def is_end_of_service(cmd: Any) -> bool:
    """
    Check if a command is the end-of-service command.
    """
    return isinstance(cmd, Command) and cmd.cmd == END_OF_SERVICE.cmd


END_OF_SERVICE = Command('stop')


# vim:sw=4:ts=4:et:



from marshmallow import pre_load

from .._base import SystemBaseSchema


class FanBaseSchema(SystemBaseSchema):
    """
    Base schema for system fan sensors.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        data['value'] = data.pop('current', data.pop('value', None))
        return data



import os
import sys
from contextlib import contextmanager
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from types import ModuleType
from typing import Any, Iterator, Sequence, Generator, Optional, List

from .modules import mock_imports


class MockObject:
    """
    Generic object that can be used to mock anything.
    """

    __display_name__ = "MockObject"
    __name__ = ""
    __decorator_args__: tuple[Any, ...] = ()

    def __new__(cls, *args: Any, **_) -> Any:
        if len(args) == 3 and isinstance(args[1], tuple):
            superclass = args[1][-1].__class__
            if superclass is cls:
                # subclassing MockObject
                return _make_subclass(
                    args[0],
                    superclass.__display_name__,
                    superclass=superclass,
                    attributes=args[2],
                )

        return super().__new__(cls)

    def __init__(self, *_, **__) -> None:
        self.__qualname__ = self.__name__

    def __len__(self) -> int:
        """
        Override __len__ so it returns zero.
        """
        return 0

    def __contains__(self, _: str) -> bool:
        """
        Override __contains__ so it always returns False.
        """
        return False

    def __iter__(self) -> Iterator:
        """
        Override __iter__ so it always returns an empty iterator.
        """
        return iter([])

    def __mro_entries__(self, _: tuple) -> tuple:
        """
        Override __mro_entries__ so it always returns a tuple containing the
        class itself.
        """
        return (self.__class__,)

    def __getitem__(self, key: Any) -> "MockObject":
        """
        Override __getitem__ so it always returns a new MockObject.
        """
        return _make_subclass(str(key), self.__display_name__, self.__class__)()

    def __getattr__(self, key: str) -> "MockObject":
        """
        Override __getattr__ so it always returns a new MockObject.
        """
        return _make_subclass(key, self.__display_name__, self.__class__)()

    def __call__(self, *args: Any, **_) -> Any:
        """
        Override __call__ so it always returns a new MockObject.
        """
        call = self.__class__()
        call.__decorator_args__ = args
        return call

    def __repr__(self) -> str:
        """
        Override __repr__ to return the display name.
        """
        return self.__display_name__


def _make_subclass(
    name: str,
    module: str,
    superclass: Any = MockObject,
    attributes: Any = None,
    decorator_args: tuple = (),
) -> Any:
    """
    Utility method that creates a mock subclass on the fly given its
    parameters.
    """
    attrs = {
        "__module__": module,
        "__display_name__": module + "." + name,
        "__name__": name,
        "__decorator_args__": decorator_args,
    }

    attrs.update(attributes or {})
    return type(name, (superclass,), attrs)


# pylint: disable=too-few-public-methods
class MockModule(ModuleType):
    """
    Object that can be used to mock any module.
    """

    __file__ = os.devnull

    def __init__(self, name: str):
        super().__init__(name)
        self.__all__ = []
        self.__path__ = []

    def __getattr__(self, name: str):
        """
        Override __getattr__ so it always returns a new MockObject.
        """
        return _make_subclass(name, self.__name__)()

    def __mro_entries__(self, _: tuple) -> tuple:
        """
        Override __mro_entries__ so it always returns a tuple containing the
        class itself.
        """
        return (self.__class__,)


class MockFinder(MetaPathFinder):
    """A finder for mocking."""

    def __init__(self, modules: Sequence[str]) -> None:  # noqa
        super().__init__()
        self.modules = modules
        self.loader = MockLoader(self)
        self.mocked_modules: List[str] = []

    def find_spec(
        self,
        fullname: str,
        path: Optional[Sequence[Optional[bytes]]] = None,
        target: Optional[ModuleType] = None,
    ) -> Optional[ModuleSpec]:
        for modname in self.modules:
            # check if fullname is (or is a descendant of) one of our targets
            if modname == fullname or fullname.startswith(modname + "."):
                return ModuleSpec(fullname, self.loader)

        return None

    def invalidate_caches(self) -> None:
        """Invalidate mocked modules on sys.modules."""
        for modname in self.mocked_modules:
            sys.modules.pop(modname, None)


class MockLoader(Loader):
    """A loader for mocking."""

    def __init__(self, finder: MockFinder) -> None:
        super().__init__()
        self.finder = finder

    def create_module(self, spec: ModuleSpec) -> ModuleType:
        self.finder.mocked_modules.append(spec.name)
        return MockModule(spec.name)

    def exec_module(self, module: ModuleType) -> None:
        pass  # nothing to do


@contextmanager
def mock(*mods: str) -> Generator[None, None, None]:
    """
    Insert mock modules during context::

    with mock('target.module.name'):
        # mock modules are enabled here
        ...
    """
    finder = None
    try:
        finder = MockFinder(mods)
        sys.meta_path.insert(0, finder)
        yield
    finally:
        if finder:
            sys.meta_path.remove(finder)
            finder.invalidate_caches()


@contextmanager
def auto_mocks():
    """
    Automatically mock all the modules listed in ``mock_imports``.
    """
    with mock(*mock_imports):
        yield


__all__ = [
    "auto_mocks",
    "mock",
]



from .auth import (
    authenticate,
    authenticate_token,
    authenticate_user_pass,
    get_auth_status,
)
from .bus import bus, send_message, send_request
from .logger import logger
from .routes import (
    get_http_port,
    get_ip_or_hostname,
    get_local_base_url,
    get_remote_base_url,
    get_routes,
)
from .streaming import get_streaming_routes
from .ws import get_ws_routes

__all__ = [
    'authenticate',
    'authenticate_token',
    'authenticate_user_pass',
    'bus',
    'get_auth_status',
    'get_http_port',
    'get_ip_or_hostname',
    'get_local_base_url',
    'get_remote_base_url',
    'get_routes',
    'get_streaming_routes',
    'get_ws_routes',
    'logger',
    'send_message',
    'send_request',
]


# vim:sw=4:ts=4:et:



# pylint: disable=too-few-public-methods

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Type

from platypush.context import get_plugin
from platypush.entities import Entity

from ._constants import DeviceType


class EntitySetter(ABC):
    """
    Base class for entity setters.

    The purpose of entity setters is to map property/values passed to
    :meth:`platypush.plugins.switchbot.SwitchbotPlugin.set_value` to native
    Switchbot device commands.
    """

    def __init__(self, entity: Entity):
        self.entity = entity
        self.device_id, self.property = self._plugin._split_device_id_and_property(
            self.entity.id
        )

    @abstractmethod
    def _set(
        self,
        value: Any,
        *args: Any,
        property: Optional[str] = None,  # pylint: disable=redefined-builtin
        **kwargs: Any,
    ):
        raise NotImplementedError()

    def __call__(
        self,
        value: Any,
        *args: Any,
        property: Optional[str] = None,  # pylint: disable=redefined-builtin
        **kwargs: Any,
    ):
        return self._set(value, *args, property=property, **kwargs)

    @property
    def _plugin(self):
        return get_plugin('switchbot')


class EntitySetterWithBinaryState(EntitySetter):
    """
    Base setter for entities with a binary on/off state.
    """

    def _set(
        self,
        value: Any,
        *_: Any,
        property: Optional[str] = None,  # pylint: disable=redefined-builtin
        **__: Any,
    ):
        if property == 'state':
            action = self._plugin.on if value else self._plugin.off
            return action(self.device_id)

        return None


class EntitySetterWithValueAsMethod(EntitySetter):
    """
    This mapper maps the value passed to
    :meth:`platypush.plugins.switchbot.SwitchbotPlugin.set_value` to plugin
    actions.

    In this case, the action value has a 1-1 mapping with the name of the
    associated plugin action.
    """

    def _set(self, value: Any, *_: Any, **__: Any):
        method = getattr(self._plugin, value, None)
        assert (
            method
        ), f'No such action available for device "{self.device_id}": "{value}"'
        return method(self.device_id)


class CurtainEntitySetter(EntitySetter):
    """
    Curtain entity setter.
    """

    def _set(self, value: Any, *_: Any, **__: Any):
        return self._plugin.set_curtain_position(self.device_id, int(value))


class HumidifierEntitySetter(EntitySetterWithBinaryState):
    """
    Humidifier entity setter.
    """

    def _set(
        self,
        value: Any,
        *args: Any,
        property: Optional[str] = None,  # pylint: disable=redefined-builtin
        **kwargs: Any,
    ):
        if property == 'state':
            return super()._set(value, *args, property=property, **kwargs)

        if property == 'child_lock':
            action = self._plugin.lock if value else self._plugin.unlock
            return action(self.device_id)

        if property in {'auto', 'nebulization_efficiency'}:
            return self._plugin.set_humidifier_efficiency(self.device_id, value)

        return None


class PlugEntitySetter(EntitySetterWithBinaryState):
    """
    Plug entity setter.
    """


class LightEntitySetter(EntitySetter):
    """
    Light entity setter.
    """

    def _set(
        self,
        value: Any,
        *_: Any,
        property: Optional[str] = None,  # pylint: disable=redefined-builtin
        **__: Any,
    ):
        assert property, 'No light property specified'
        return self._plugin.set_curtain_position(self.device_id, int(value))


# A static map of device types -> entity setters functors.
entity_setters: Dict[DeviceType, Type[EntitySetter]] = {
    DeviceType.BOT: EntitySetterWithValueAsMethod,
    DeviceType.CEILING_LIGHT: LightEntitySetter,
    DeviceType.CEILING_LIGHT_PRO: LightEntitySetter,
    DeviceType.COLOR_BULB: LightEntitySetter,
    DeviceType.CURTAIN: CurtainEntitySetter,
    DeviceType.HUMIDIFIER: HumidifierEntitySetter,
    DeviceType.LOCK: EntitySetterWithValueAsMethod,
    DeviceType.PLUG: PlugEntitySetter,
    DeviceType.PLUG_MINI_US: PlugEntitySetter,
    DeviceType.PLUG_MINI_JP: PlugEntitySetter,
    DeviceType.STRIP_LIGHT: LightEntitySetter,
}



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Iterable, List, Tuple

from sqlalchemy import and_, or_
from sqlalchemy.orm import Session

from platypush.context import get_plugin
from platypush.entities._base import Entity
from .helpers import get_parent


@dataclass
class _TaxonomyAwareEntity:
    """
    A support class used to map an entity and its level within a taxonomy to be
    flushed to the database.
    """

    entity: Entity
    level: int


class EntitiesDb:
    """
    This object is a facade around the entities database. It shouldn't be used
    directly. Instead, it is encapsulated by
    :class:`platypush.entities._repo.EntitiesRepository`, which is in charge of
    caching as well.
    """

    def get_session(self, *args, **kwargs) -> Session:
        db = get_plugin('db')
        assert db
        return db.get_session(*args, **kwargs)

    def fetch(
        self, session: Session, entities: Iterable[Entity]
    ) -> Dict[Tuple[str, str], Entity]:
        """
        Given a set of entities, it returns those that already exist on the database.
        """
        if not entities:
            return {}

        entities_filter = or_(
            *[
                and_(
                    Entity.external_id == entity.external_id,
                    Entity.plugin == entity.plugin,
                )
                for entity in entities
            ]
        )

        query = session.query(Entity).filter(entities_filter)
        existing_entities = {entity.entity_key: entity for entity in query.all()}

        return {
            entity.entity_key: existing_entities[entity.entity_key]
            for entity in entities
            if existing_entities.get(entity.entity_key)
        }

    @staticmethod
    def _close_batch(batch: List[_TaxonomyAwareEntity], batches: List[List[Entity]]):
        if batch:
            batches.append([item.entity for item in batch])

            batch.clear()

    def _split_entity_batches_for_flush(
        self, session: Session, entities: Iterable[Entity]
    ) -> List[List[Entity]]:
        """
        This method retrieves the root entities given a list of entities and
        generates batches of "flushable" entities ready for upsert using a BFS
        algorithm.

        This is needed because we want hierarchies of entities to be flushed
        starting from the top layer, once their parents have been appropriately
        rewired. Otherwise, we may end up with conflicts on entities that have
        already been flushed.
        """
        # Index children by parent_id and by parent_key
        children_by_parent_id: Dict[int, Dict[Tuple[str, str], Entity]] = defaultdict(
            lambda: defaultdict(Entity)
        )

        children_by_parent_key: Dict[
            Tuple[str, str], Dict[Tuple[str, str], Entity]
        ] = defaultdict(lambda: defaultdict(Entity))

        for entity in entities:
            parent_key = None
            parent_id = entity.parent_id
            parent = get_parent(session, entity)
            if parent:
                parent_id = parent_id or parent.id
                parent_key = parent.entity_key

            if parent_id:
                children_by_parent_id[parent_id][entity.entity_key] = entity
            if parent_key:
                children_by_parent_key[parent_key][entity.entity_key] = entity

        # Find the root entities in the hierarchy (i.e. those that have a null
        # parent)
        root_entities = list(
            {
                e.entity_key: e
                for e in entities
                if e.parent is None and e.parent_id is None
            }.values()
        )

        # Prepare a list of entities to process through BFS starting with the
        # root nodes (i.e. level=0)
        entities_to_process = [
            _TaxonomyAwareEntity(entity=e, level=0) for e in root_entities
        ]

        batches: List[List[Entity]] = []
        current_batch: List[_TaxonomyAwareEntity] = []

        while entities_to_process:
            # Pop the first element in the list (FIFO implementation)
            item = entities_to_process.pop(0)
            entity = item.entity
            level = item.level

            # If the depth has increased compared to the previous node, flush
            # the current batch and open a new one.
            if current_batch and current_batch[-1].level < level:
                self._close_batch(current_batch, batches)
            current_batch.append(item)

            # Index the children nodes by key
            children_to_process = {
                e.entity_key: e
                for e in children_by_parent_key.get(entity.entity_key, {}).values()
            }

            # If this entity has already been persisted, add back its children
            # that haven't been updated, so we won't lose those connections
            if entity.id:
                children_to_process.update(
                    {
                        e.entity_key: e
                        for e in children_by_parent_id.get(entity.id, {}).values()
                    }
                )

            # Add all the updated+inserted+existing children to the next layer
            # to be expanded
            entities_to_process += [
                _TaxonomyAwareEntity(entity=e, level=level + 1)
                for e in children_to_process.values()
            ]

        # Close any pending batches
        self._close_batch(current_batch, batches)
        return batches

    def upsert(
        self,
        session: Session,
        entities: Iterable[Entity],
    ) -> Iterable[Entity]:
        """
        Persist a set of entities.
        """
        # Get the "unwrapped" batches
        batches = self._split_entity_batches_for_flush(session, entities)

        # Flush each batch as we process it
        for batch in batches:
            session.add_all(batch)
            session.flush()

        session.commit()

        # Make a copy of the entities in the batch, so they can be used outside
        # of this session/thread without the DetachedInstanceError pain
        return list(
            {
                entity.entity_key: entity.copy()
                for batch in batches
                for entity in batch
            }.values()
        )



import datetime
import glob
import importlib
import inspect
import json
import logging
import os
import pathlib
import pkgutil
import re
import shutil
import socket
import sys
from urllib.parse import quote
from typing import Any, Dict, Optional, Set

import yaml

from platypush.utils import (
    get_hash,
    is_functional_procedure,
    is_functional_hook,
    is_functional_cron,
    is_root,
)


class Config:
    """
    Configuration base class
    Usage:
        - Initialize config from one of the default paths:
            Config.init()
        - Initialize config from a custom path
            Config.init(config_file_path)
        - Get a value
            Config.get('foo')
    """

    # Default config file locations:
    #   - current directory
    #   - $XDG_CONFIG_HOME/platypush/config.yaml
    #   - $HOME/.config/platypush/config.yaml
    #   - /etc/platypush/config.yaml
    _cfgfile_locations = [
        os.path.join(os.path.abspath('.'), 'config.yaml'),
        os.path.join(os.environ.get('XDG_CONFIG_HOME', ''), 'config.yaml'),
        os.path.join(os.path.expanduser('~'), '.config', 'platypush', 'config.yaml'),
        os.path.join(os.sep, 'etc', 'platypush', 'config.yaml'),
    ]

    # Config singleton instance
    _instance = None

    _default_constants = {
        'today': datetime.date.today,
        'now': datetime.datetime.now,
    }

    # Default working directory:
    #  - $XDG_DATA_HOME/platypush if XDG_DATA_HOME is set
    #  - /var/lib/platypush if the user is root
    #  - $HOME/.local/share/platypush otherwise
    _workdir_location = os.path.join(
        *(
            (os.environ['XDG_DATA_HOME'],)
            if os.environ.get('XDG_DATA_HOME')
            else (
                (os.sep, 'var', 'lib')
                if os.geteuid() == 0
                else (os.path.expanduser('~'), '.local', 'share')
            )
        ),
        'platypush',
    )

    # Default cache directory:
    #  - $XDG_CACHE_DIR/platypush if XDG_CACHE_DIR is set
    #  - /var/cache/platypush if the user is root
    #  - $HOME/.cache/platypush otherwise
    _cachedir_location = os.path.join(
        *(
            (os.environ['XDG_CACHE_DIR'],)
            if os.environ.get('XDG_CACHE_DIR')
            else (
                (os.sep, 'var', 'cache')
                if os.geteuid() == 0
                else (os.path.expanduser('~'), '.cache')
            )
        ),
        'platypush',
    )

    _included_files: Set[str] = set()

    def __init__(
        self,
        cfgfile: Optional[str] = None,
        workdir: Optional[str] = None,
        cachedir: Optional[str] = None,
        db: Optional[str] = None,
    ):
        """
        Constructor. Always use the class as a singleton (i.e. through
        Config.init), you won't probably need to call the constructor directly

        :param cfgfile: Config file path (default: retrieve the first available
            location in _cfgfile_locations).
        :param workdir: Overrides the default working directory.
        :param cachedir: Overrides the default cache directory.
        :param db: Overrides the default database connection string.
        """

        self.backends = {}
        self.plugins = self._core_plugins
        self.event_hooks = {}
        self.procedures = {}
        self.constants = {}
        self.cronjobs = {}
        self.dashboards = {}
        self._plugin_manifests = {}
        self._backend_manifests = {}
        self.config_file = ''

        self._init_cfgfile(cfgfile)
        self._config = self._read_config_file(self.config_file)

        self._init_secrets()
        self._init_dirs(workdir=workdir, cachedir=cachedir)
        self._init_db(db=db)
        self._init_logging()
        self._init_device_id()
        self._init_environment()
        self._init_manifests()
        self._init_constants()
        self._load_scripts()
        self._init_components()
        self._init_dashboards(self._config['dashboards_dir'])

    def _init_cfgfile(self, cfgfile: Optional[str] = None):
        if cfgfile is None:
            cfgfile = self._get_default_cfgfile()

        if cfgfile:
            cfgfile = os.path.abspath(os.path.expanduser(cfgfile))

        if cfgfile is None or not os.path.exists(cfgfile):
            cfgfile = self._create_default_config(cfgfile)

        self.config_file = cfgfile

    def _init_logging(self):
        logging_config = {
            'level': logging.INFO,
            'stream': sys.stdout,
            'format': '%(asctime)-15s|%(levelname)5s|%(name)s|%(message)s',
        }

        if 'logging' in self._config:
            for k, v in self._config['logging'].items():
                if k == 'filename':
                    logfile = os.path.expanduser(v)
                    logdir = os.path.dirname(logfile)
                    try:
                        os.makedirs(logdir, exist_ok=True)
                    except Exception as e:
                        print(f'Unable to create logs directory {logdir}: {e}')

                    v = logfile
                    del logging_config['stream']
                elif k == 'level':
                    try:
                        v = int(v)
                    except ValueError:
                        v = getattr(logging, v.upper())

                logging_config[k] = v

        self._config['logging'] = logging_config

    def _init_db(self, db: Optional[str] = None):
        # If the db connection string is passed as an argument, use it
        if db:
            self._config['db'] = {
                'engine': db,
            }
            return

        # Initialize the default db connection string
        db_engine = self._config.get('main.db', '')
        if db_engine:
            if isinstance(db_engine, str):
                db_engine = {
                    'engine': db_engine,
                }
        else:
            db_engine = {
                'engine': 'sqlite:///'
                + os.path.join(quote(self._config['workdir']), 'main.db')
            }

        self._config['db'] = db_engine

    def _init_device_id(self):
        if 'device_id' not in self._config:
            self._config['device_id'] = socket.gethostname()

    def _init_environment(self):
        if 'environment' in self._config:
            for k, v in self._config['environment'].items():
                os.environ[k] = str(v)

    def _init_workdir(self, workdir: Optional[str] = None):
        if workdir:
            self._config['workdir'] = workdir
        if not self._config.get('workdir'):
            self._config['workdir'] = self._workdir_location

        self._config['workdir'] = os.path.expanduser(self._config['workdir'])
        pathlib.Path(self._config['workdir']).mkdir(parents=True, exist_ok=True)

    def _init_cachedir(self, cachedir: Optional[str] = None):
        if cachedir:
            self._config['cachedir'] = cachedir
        if not self._config.get('cachedir'):
            self._config['cachedir'] = self._cachedir_location

        self._config['cachedir'] = os.path.expanduser(self._config['cachedir'])
        pathlib.Path(self._config['cachedir']).mkdir(parents=True, exist_ok=True)

    def _init_scripts_dir(self):
        # Create the scripts directory if it doesn't exist
        if 'scripts_dir' not in self._config:
            self._config['scripts_dir'] = os.path.join(
                os.path.dirname(self.config_file), 'scripts'
            )

        scripts_dir = os.path.abspath(os.path.expanduser(self._config['scripts_dir']))
        os.makedirs(scripts_dir, mode=0o755, exist_ok=True)

        # Make sure that every folder with a .py file has an __init__.py file
        for root, _, files in os.walk(scripts_dir):
            for file in files:
                if file.endswith('.py'):
                    init_py = os.path.join(root, '__init__.py')
                    if not os.path.isfile(init_py):
                        with open(init_py, 'w') as f:
                            f.write('')

        # Include scripts_dir parent in sys.path so members can be imported in scripts
        # through the `scripts` package
        scripts_parent_dir = str(pathlib.Path(scripts_dir).absolute().parent)
        sys.path = [scripts_parent_dir] + sys.path

    def _init_dashboards_dir(self):
        if 'dashboards_dir' not in self._config:
            self._config['dashboards_dir'] = os.path.join(
                os.path.dirname(self.config_file), 'dashboards'
            )
        os.makedirs(self._config['dashboards_dir'], mode=0o755, exist_ok=True)

    def _init_dirs(self, workdir: Optional[str] = None, cachedir: Optional[str] = None):
        self._init_workdir(workdir=workdir)
        self._init_cachedir(cachedir=cachedir)
        self._init_scripts_dir()
        self._init_dashboards_dir()

    def _init_secrets(self):
        if 'token' in self._config:
            self._config['token_hash'] = get_hash(self._config['token'])

    @property
    def _core_plugins(self) -> Dict[str, dict]:
        return {
            'variable': {},
        }

    @staticmethod
    def _create_default_config(cfgfile: Optional[str] = None):
        cfg_mod_dir = os.path.dirname(os.path.abspath(__file__))

        if not cfgfile:
            # Use /etc/platypush/config.yaml if the user is running as root,
            # otherwise ~/.config/platypush/config.yaml
            cfgfile = (
                (
                    os.path.join(os.environ['XDG_CONFIG_HOME'], 'config.yaml')
                    if os.environ.get('XDG_CONFIG_HOME')
                    else os.path.join(
                        os.path.expanduser('~'), '.config', 'platypush', 'config.yaml'
                    )
                )
                if not is_root()
                else os.path.join(os.sep, 'etc', 'platypush', 'config.yaml')
            )

        cfgdir = pathlib.Path(cfgfile).parent
        cfgdir.mkdir(parents=True, exist_ok=True)
        for cf in glob.glob(os.path.join(cfg_mod_dir, 'config*.yaml')):
            shutil.copy(cf, str(cfgdir))

        return cfgfile

    def _read_config_file(self, cfgfile):
        cfgfile_dir = os.path.dirname(os.path.abspath(os.path.expanduser(cfgfile)))
        config = {}

        try:
            with open(cfgfile, 'r') as fp:
                file_config = yaml.safe_load(fp)
        except FileNotFoundError:
            print(f'Unable to open config file {cfgfile}')
            return config

        if not file_config:
            return config

        for section in file_config:
            if section == 'include':
                include_files = (
                    file_config[section]
                    if isinstance(file_config[section], list)
                    else [file_config[section]]
                )

                for include_file in include_files:
                    if not include_file:
                        continue
                    if not os.path.isabs(include_file):
                        include_file = os.path.join(cfgfile_dir, include_file)

                    self._included_files.add(include_file)
                    config.update(self._read_config_file(include_file))
            elif section == 'scripts_dir':
                assert isinstance(file_config[section], str)
                config['scripts_dir'] = os.path.abspath(
                    os.path.expanduser(file_config[section])
                )
            else:
                section_config = file_config.get(section, {}) or {}
                if not (
                    hasattr(section_config, 'get') and section_config.get('disabled')
                ):
                    config[section] = section_config

        return config

    def _load_module(self, modname: str, prefix: Optional[str] = None):
        try:
            module = importlib.import_module(modname)
        except Exception as e:
            print(f'Unhandled exception while importing module {modname}: {e}')
            return

        prefix = modname + '.' if prefix is None else prefix
        self.procedures.update(
            **{
                (getattr(obj, 'procedure_name', None) or f'{prefix}{name}'): obj
                for name, obj in inspect.getmembers(module)
                if is_functional_procedure(obj)
            }
        )

        self.event_hooks.update(
            **{
                prefix + name: obj
                for name, obj in inspect.getmembers(module)
                if is_functional_hook(obj)
            }
        )

        self.cronjobs.update(
            **{
                prefix + name: obj
                for name, obj in inspect.getmembers(module)
                if is_functional_cron(obj)
            }
        )

    def _load_scripts(self):
        scripts_dir = self._config['scripts_dir']
        sys_path = sys.path.copy()
        sys.path = [scripts_dir] + sys.path
        scripts_modname = os.path.basename(scripts_dir)
        self._load_module(scripts_modname, prefix='')

        for _, modname, _ in pkgutil.walk_packages(
            path=[scripts_dir], onerror=lambda _: None
        ):
            self._load_module(modname)

        sys.path = sys_path

    def _init_components(self):
        for key, component in self._config.items():
            if (
                key.startswith('backend.')
                and '.'.join(key.split('.')[1:]) in self._backend_manifests
            ):
                backend_name = '.'.join(key.split('.')[1:])
                self.backends[backend_name] = component
            elif key.startswith('event.hook.'):
                hook_name = '.'.join(key.split('.')[2:])
                self.event_hooks[hook_name] = component
            elif key.startswith('cron.'):
                cron_name = '.'.join(key.split('.')[1:])
                self.cronjobs[cron_name] = component
            elif key.startswith('procedure.'):
                tokens = key.split('.')
                _async = bool(len(tokens) > 2 and tokens[1] == 'async')
                procedure_name = '.'.join(tokens[2:] if len(tokens) > 2 else tokens[1:])
                args = []
                m = re.match(r'^([^(]+)\(([^)]+)\)\s*', procedure_name)

                if m:
                    procedure_name = m.group(1).strip()
                    args = [
                        arg.strip()
                        for arg in m.group(2).strip().split(',')
                        if arg.strip()
                    ]

                self.procedures[procedure_name] = {
                    '_async': _async,
                    'actions': component,
                    'args': args,
                }
            elif key in self._plugin_manifests:
                self.plugins[key] = component

    def _init_manifests(self, base_dir: Optional[str] = None):
        if not base_dir:
            base_dir = os.path.abspath(os.path.join(__file__, '..', '..'))
            plugins_dir = os.path.join(base_dir, 'plugins')
            backends_dir = os.path.join(base_dir, 'backend')
            self._init_manifests(plugins_dir)
            self._init_manifests(backends_dir)
        else:
            manifests_map = (
                self._plugin_manifests
                if base_dir.endswith('plugins')
                else self._backend_manifests
            )

            for mf in pathlib.Path(base_dir).rglob('manifest.json'):
                with open(mf, 'r') as f:
                    manifest = json.load(f).get('manifest')
                    if not manifest:
                        continue

                    comp_name = '.'.join(manifest['package'].split('.')[2:])
                    manifests_map[comp_name] = manifest

    def _init_constants(self):
        if 'constants' in self._config:
            self.constants = self._config['constants']

        for key, value in self._default_constants.items():
            self.constants[key] = value

    def _get_dashboard(
        self, name: str, dashboards_dir: Optional[str] = None
    ) -> Optional[str]:
        dashboards_dir = dashboards_dir or self._config['dashboards_dir']
        assert dashboards_dir
        abspath = os.path.join(dashboards_dir, name + '.xml')
        if not os.path.isfile(abspath):
            return None

        with open(abspath, 'r') as fp:
            return fp.read()

    def _get_dashboards(self, dashboards_dir: Optional[str] = None) -> dict:
        dashboards = {}
        dashboards_dir = dashboards_dir or self._config['dashboards_dir']
        assert dashboards_dir

        for f in os.listdir(dashboards_dir):
            abspath = os.path.join(dashboards_dir, f)
            if not os.path.isfile(abspath) or not abspath.endswith('.xml'):
                continue

            name = f.split('.xml')[0]
            dashboards[name] = self._get_dashboard(name, dashboards_dir)

        return dashboards

    @classmethod
    def _get_instance(
        cls,
        cfgfile: Optional[str] = None,
        workdir: Optional[str] = None,
        cachedir: Optional[str] = None,
        db: Optional[str] = None,
        force_reload: bool = False,
    ) -> "Config":
        """
        Lazy getter/setter for the default configuration instance.
        """
        if force_reload or cls._instance is None:
            cfg_args = [cfgfile] if cfgfile else []
            cls._instance = Config(*cfg_args, workdir=workdir, cachedir=cachedir, db=db)
        return cls._instance

    @classmethod
    def get_dashboard(
        cls, name: str, dashboards_dir: Optional[str] = None
    ) -> Optional[str]:
        # pylint: disable=protected-access
        return cls._get_instance()._get_dashboard(name, dashboards_dir)

    @classmethod
    def get_dashboards(cls, dashboards_dir: Optional[str] = None) -> dict:
        # pylint: disable=protected-access
        return cls._get_instance()._get_dashboards(dashboards_dir)

    def _init_dashboards(self, dashboards_dir: str):
        self.dashboards = self._get_dashboards(dashboards_dir)

    @classmethod
    def get_backends(cls):
        return cls._get_instance().backends

    @classmethod
    def get_plugins(cls):
        return cls._get_instance().plugins

    @classmethod
    def get_event_hooks(cls):
        return cls._get_instance().event_hooks

    @classmethod
    def get_procedures(cls):
        return cls._get_instance().procedures

    @classmethod
    def get_constants(cls):
        return {
            name: Config.get_constant(name) for name in cls._get_instance().constants
        }

    @classmethod
    def get_constant(cls, name):
        value = cls._get_instance().constants.get(name)
        if value is None:
            return None
        return value() if callable(value) else value

    @classmethod
    def get_cronjobs(cls):
        return cls._get_instance().cronjobs

    @classmethod
    def _get_default_cfgfile(cls) -> Optional[str]:
        for location in cls._cfgfile_locations:
            if os.path.isfile(location):
                return location
        return None

    @classmethod
    def init(
        cls,
        cfgfile: Optional[str] = None,
        device_id: Optional[str] = None,
        workdir: Optional[str] = None,
        cachedir: Optional[str] = None,
        db: Optional[str] = None,
        ctrl_sock: Optional[str] = None,
        **_,
    ):
        """
        Initializes the config object singleton

        :param cfgfile: Path to the config file (default: _cfgfile_locations)
        :param device_id: Override the configured device_id.
        :param workdir: Override the configured working directory.
        :param cachedir: Override the configured cache directory.
        :param db: Override the configured database connection string.
        :param ctrl_sock: Override the configured control socket.
        """
        cfg = cls._get_instance(
            cfgfile, workdir=workdir, cachedir=cachedir, db=db, force_reload=True
        )
        if device_id:
            cfg.set('device_id', device_id)
        if workdir:
            cfg.set('workdir', workdir)
        if cachedir:
            cfg.set('cachedir', cachedir)
        if db:
            cfg.set('db', db)
        if ctrl_sock:
            cfg.set('ctrl_sock', ctrl_sock)

        return cfg

    @classmethod
    def get_workdir(cls) -> str:
        """
        :return: The path of the configured working directory.
        """
        workdir = cls._get_instance().get('workdir')
        assert workdir
        return workdir  # type: ignore

    @classmethod
    def get_device_id(cls) -> str:
        """
        :return: The configured/default device ID.
        """
        device_id = cls._get_instance().get('device_id')
        assert device_id
        return device_id  # type: ignore

    @classmethod
    def get_cachedir(cls) -> str:
        """
        :return: The path of the configured cache directory.
        """
        workdir = cls._get_instance().get('cachedir')
        assert workdir
        return workdir  # type: ignore

    @classmethod
    def get(cls, key: Optional[str] = None, default: Optional[Any] = None):
        """
        Get a config value or the whole configuration object.

        :param key: Configuration entry to get (default: all entries).
        :param default: Default value to return if the key is missing.
        """
        # pylint: disable=protected-access
        config = cls._get_instance()._config.copy()
        if key:
            return config.get(key, default)
        return config

    @classmethod
    def set(cls, key: str, value: Any):
        """
        Set a config value.

        :param key: Configuration entry to set.
        :param value: New value to set.
        """
        # pylint: disable=protected-access
        cls._get_instance()._config[key] = value

    @classmethod
    def get_file(cls) -> str:
        """
        :return: The main configuration file path.
        """
        return cls._get_instance().config_file


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from socket import AddressFamily, SocketKind
from typing import Optional


@dataclass
class Connection:
    """
    Network/UNIX socket data class.
    """

    fd: int = field(
        metadata={
            'metadata': {
                'description': 'File descriptor',
                'example': 3,
            },
        }
    )

    family: AddressFamily = field(
        metadata={
            'metadata': {
                'description': 'Socket family',
                'example': AddressFamily.AF_INET.name,
            }
        }
    )

    type: SocketKind = field(
        metadata={
            'metadata': {
                'description': 'Socket type',
                'example': SocketKind.SOCK_STREAM.name,
            }
        }
    )

    local_address: str = field(
        metadata={
            'metadata': {
                'description': 'Local address, as an IP address for network '
                'connections and a socket path for a UNIX socket',
                'example': '192.168.1.2',
            }
        }
    )

    local_port: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'Local port, if this is a TCP/UDP connection, '
                'otherwise null',
                'example': 12345,
            }
        }
    )

    remote_address: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Remote address, if this is a network '
                'connection, otherwise null',
                'example': '192.168.1.1',
            }
        }
    )

    remote_port: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'Local port, if this is a TCP/UDP connection, '
                'otherwise null',
                'example': 443,
            }
        }
    )

    status: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Connection status, if this is a network '
                'connection, otherise null',
                'example': 'ESTABLISHED',
            }
        }
    )

    pid: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'ID of the process that owns the connection',
                'example': 4321,
            }
        }
    )



from datetime import datetime as dt
from typing import Optional

from platypush.message.event import Event


class NextCloudActivityEvent(Event):
    """
    Event triggered when a new activity is detected on a NextCloud instance.
    """

    def __init__(
        self,
        *args,
        activity_id: int,
        activity_type: str,
        object_id: int,
        object_type: str,
        object_name: str,
        app: str,
        user: str,
        subject: str,
        message: str,
        subject_rich: Optional[list] = None,
        message_rich: Optional[list] = None,
        objects: Optional[dict] = None,
        link: Optional[str] = None,
        icon: Optional[str] = None,
        datetime: Optional[dt] = None,
        **kwargs,
    ):
        """
        :param activity_id: Activity ID.
        :param activity_type: Activity type - can be ``file_created``,
            ``file_deleted``, ``file_changed``, ``file_restored``,
            ``file_shared``, ``file_unshared``, ``file_downloaded``, etc.
        :param object_id: Object ID.
        :param object_type: Object type - can be files, comment, tag, share,
            etc.
        :param object_name: Object name. In the case of files, it's the file
            path relative to the user's root directory.
        :param app: Application that generated the activity.
        :param user: User that generated the activity.
        :param subject: Activity subject, in plain text. For example, *You
            created hd/test1.txt and hd/test2.txt*.
        :param message: Activity message, in plain text.
        :param subject_rich: Activity subject, in rich/structured format.
            Example:

              .. code-block:: json

                [
                  "You created {file2} and {file1}",
                  {
                    "file1": {
                      "type": "file",
                      "id": "1234",
                      "name": "test1.txt",
                      "path": "hd/text1.txt",
                      "link": "https://cloud.example.com/index.php/f/1234"
                    },
                    "file2": {
                      "type": "file",
                      "id": "1235",
                      "name": "test2.txt",
                      "path": "hd/text2.txt",
                      "link": "https://cloud.example.com/index.php/f/1235"
                    }
                  }
                ]

        :param message_rich: Activity message, in rich/structured format.
        :param objects: Additional objects associated to the activity, in the
            format ``{object_id: object}``. For example, if the activity
            involves files, the ``objects`` dictionary will contain the mapping
            of the involved files in the format ``{file_id: path}``.
        :param link: Link to the main object of this activity. Example:
            ``https://cloud.example.com/index.php/files/apps/files/?dir=/hd&fileid=1234``
        :param icon: URL of the icon associated to the activity.
        :param datetime: Activity timestamp.
        """
        super().__init__(
            *args,
            activity_id=activity_id,
            activity_type=activity_type,
            object_id=object_id,
            object_type=object_type,
            object_name=object_name,
            app=app,
            user=user,
            subject=subject,
            subject_rich=subject_rich,
            message=message,
            message_rich=message_rich,
            objects=objects or {},
            link=link,
            icon=icon,
            datetime=datetime,
            **kwargs,
        )


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from threading import Thread
from typing import Optional

from platypush.context import get_bus
from platypush.message.event.music import (
    MusicPlayEvent,
    MusicPauseEvent,
    MusicStopEvent,
    NewPlayingTrackEvent,
    PlaylistChangeEvent,
    VolumeChangeEvent,
    PlaybackConsumeModeChangeEvent,
    PlaybackSingleModeChangeEvent,
    PlaybackRepeatModeChangeEvent,
    PlaybackRandomModeChangeEvent,
)
from platypush.plugins.music import MusicPlugin


@dataclass
class MpdStatus:
    """
    Data class for the MPD status.
    """

    state: Optional[str] = None
    playlist: Optional[int] = None
    volume: Optional[int] = None
    random: Optional[bool] = None
    repeat: Optional[bool] = None
    consume: Optional[bool] = None
    single: Optional[bool] = None
    track: dict = field(default_factory=dict)


class MpdListener(Thread):
    """
    Thread that listens/polls for MPD events and posts them to the bus.
    """

    def __init__(self, plugin: MusicPlugin, *_, **__):
        from . import MusicMpdPlugin

        super().__init__(name='platypush:mpd:listener')
        assert isinstance(plugin, MusicMpdPlugin)
        self.plugin: MusicMpdPlugin = plugin
        self._status = MpdStatus()

    @property
    def logger(self):
        return self.plugin.logger

    @property
    def bus(self):
        return get_bus()

    def wait_stop(self, timeout=None):
        self.plugin.wait_stop(timeout=timeout)

    def _process_events(self, status: dict, track: Optional[dict] = None):
        state = status.get('state', '').lower()
        evt_args = {'status': status, 'track': track, 'plugin_name': 'music.mpd'}

        if state != self._status.state:
            if state == 'stop':
                self.bus.post(MusicStopEvent(**evt_args))
            elif state == 'pause':
                self.bus.post(MusicPauseEvent(**evt_args))
            elif state == 'play':
                self.bus.post(MusicPlayEvent(**evt_args))

        if status.get('playlist') != self._status.playlist and self._status.playlist:
            # XXX plchanges can become heavy with big playlists,
            # PlaylistChangeEvent temporarily disabled
            # changes = plugin.plchanges(last_playlist).output
            # self.bus.post(PlaylistChangeEvent(changes=changes))
            self.bus.post(PlaylistChangeEvent(plugin_name='music.mpd'))

        if state == 'play' and track != self._status.track:
            self.bus.post(NewPlayingTrackEvent(**evt_args))

        if (
            status.get('volume') is not None
            and status.get('volume') != self._status.volume
        ):
            self.bus.post(VolumeChangeEvent(volume=int(status['volume']), **evt_args))

        if (
            status.get('random') is not None
            and status.get('random') != self._status.random
        ):
            self.bus.post(
                PlaybackRandomModeChangeEvent(
                    state=bool(int(status['random'])), **evt_args
                )
            )

        if (
            status.get('repeat') is not None
            and status.get('repeat') != self._status.repeat
        ):
            self.bus.post(
                PlaybackRepeatModeChangeEvent(
                    state=bool(int(status['repeat'])), **evt_args
                )
            )

        if (
            status.get('consume') is not None
            and status.get('consume') != self._status.consume
        ):
            self.bus.post(
                PlaybackConsumeModeChangeEvent(
                    state=bool(int(status['consume'])), **evt_args
                )
            )

        if (
            status.get('single') is not None
            and status.get('single') != self._status.single
        ):
            self.bus.post(
                PlaybackSingleModeChangeEvent(
                    state=bool(int(status['single'])), **evt_args
                )
            )

    def _update_status(self, status: dict, track: Optional[dict] = None):
        self._status = MpdStatus(
            state=status.get('state', '').lower(),
            playlist=status.get('playlist'),
            volume=status.get('volume'),
            random=status.get('random'),
            repeat=status.get('repeat'),
            consume=status.get('consume'),
            single=status.get('single'),
            track=track or {},
        )

    def run(self):
        super().run()

        while not self.plugin.should_stop():
            try:
                status = self.plugin._status()  # pylint: disable=protected-access
                assert status and status.get('state'), 'No status returned'
                if not (status and status.get('state')):
                    self.wait_stop(self.plugin.poll_interval)
                    break

                track = self.plugin._current_track()  # pylint: disable=protected-access
                self._process_events(status, track)
                self._update_status(status, track)
            except Exception as e:
                self.logger.warning(
                    'Could not retrieve the latest status: %s', e, exc_info=True
                )
            finally:
                self.wait_stop(self.plugin.poll_interval)


# vim:sw=4:ts=4:et:



from typing import Iterable, Mapping, Optional, Union
from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined
from platypush.common.sensors import Numeric

from .sensors import RawSensor


if not is_defined('three_axis_sensor'):

    class ThreeAxisSensor(RawSensor):
        """
        An entity that measures a time duration.
        """

        __tablename__ = 'three_axis_sensor'

        id = Column(
            Integer, ForeignKey(RawSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }

        @RawSensor.value.setter
        def value(
            self, value: Optional[Union[Iterable[Numeric], Mapping[str, Numeric]]]
        ):
            """
            Validates and normalizes the given value to a list of 3 numeric
            values.
            """
            if value is None:
                return

            if isinstance(value, dict):
                assert set(value.keys()) == {
                    'x',
                    'y',
                    'z',
                }, f'Invalid keys for entity of type {self.__class__}: "{value}"'

                value = [value[k] for k in ['x', 'y', 'z']]  # type: ignore

            assert (
                isinstance(value, (list, tuple))
                and len(value) == 3  # type: ignore
                and all(isinstance(v, (int, float)) for v in value)
            ), f'Invalid 3-axis value: {value}'

            super(ThreeAxisSensor, type(self)).value.fset(self, value)  # type: ignore



from abc import ABC
from dataclasses import asdict, dataclass
from enum import Enum
import os
from threading import Event
from typing import Any, Collection, Dict, Optional, Type

from platypush.context import get_bus, get_plugin
from platypush.entities.assistants import Assistant
from platypush.entities.managers.assistants import AssistantEntityManager
from platypush.message.event import Event as AppEvent
from platypush.plugins import Plugin, action
from platypush.utils import get_plugin_name_by_class


class AlertType(Enum):
    """
    Enum representing the type of an alert.
    """

    ALARM = 'alarm'
    TIMER = 'timer'
    ALERT = 'alert'


@dataclass
class AssistantState:
    """
    Dataclass representing the state of an assistant.
    """

    last_query: Optional[str] = None
    last_response: Optional[str] = None
    conversation_running: bool = False
    is_muted: bool = False
    is_detecting: bool = True
    alert_state: Optional[str] = None


class AssistantPlugin(Plugin, AssistantEntityManager, ABC):
    """
    Base class for assistant plugins.
    """

    _entity_name = 'Assistant'

    def __init__(
        self,
        *args,
        tts_plugin: Optional[str] = None,
        tts_plugin_args: Optional[Dict[str, Any]] = None,
        conversation_start_sound: Optional[str] = None,
        stop_conversation_on_speech_match: bool = True,
        **kwargs,
    ):
        """
        :param tts_plugin: If set, the assistant will use this plugin (e.g.
            ``tts``, ``tts.google`` or ``tts.mimic3``) to render the responses,
            instead of using the built-in assistant voice.

        :param tts_plugin_args: Optional arguments to be passed to the TTS
            ``say`` action, if ``tts_plugin`` is set.

        :param conversation_start_sound: If set, the assistant will play this
            audio file when it detects a speech. The sound file will be played
            on the default audio output device. If not set, the assistant won't
            play any sound when it detects a speech.

        :param stop_conversation_on_speech_match: If set, the plugin will
            prevent the default assistant response when a
            :class:`platypush.message.event.assistant.SpeechRecognizedEvent`
            matches a user hook with a condition on a ``phrase`` field. This is
            useful to prevent the assistant from responding with a default "*I'm
            sorry, I can't help you with that*" when e.g. you say "*play the
            music*", and you have a hook that matches the phrase "*play the
            music*" and handles it with a custom action. If set, and you wish
            the assistant to also provide an answer if an event matches one of
            your hooks, then you should call the :meth:`render_response` method
            in your hook handler. If not set, then the assistant will always try
            and respond with a default message, even if a speech event matches
            the phrase of one of your hooks. In this case, if you want to prevent
            the default response, you should call :meth:`stop_conversation`
            explicitly from your hook handler. Default: True.
        """
        super().__init__(*args, **kwargs)
        self.tts_plugin = tts_plugin
        self.tts_plugin_args = {'join': True, **(tts_plugin_args or {})}
        self.stop_conversation_on_speech_match = stop_conversation_on_speech_match
        self._conversation_start_sound = None
        if conversation_start_sound:
            self._conversation_start_sound = os.path.abspath(
                os.path.expanduser(conversation_start_sound)
            )

        self._detection_paused = Event()
        self._conversation_running = Event()
        self._is_muted = False
        self._last_query: Optional[str] = None
        self._last_response: Optional[str] = None
        self._cur_alert_type: Optional[AlertType] = None
        self._plugin_name = get_plugin_name_by_class(type(self))

    @property
    def _state(self) -> AssistantState:
        return AssistantState(
            last_query=self._last_query,
            last_response=self._last_response,
            conversation_running=self._conversation_running.is_set(),
            is_muted=self._is_muted,
            is_detecting=not self._detection_paused.is_set(),
            alert_state=self._cur_alert_type.value if self._cur_alert_type else None,
        )

    @action
    def stop_conversation(self, *_, **__):
        """
        Programmatically stops a conversation.
        """
        self._stop_conversation()

    def _stop_conversation(self, *_, **__):
        tts = self._get_tts_plugin()
        if tts:
            tts.stop()

    @action
    def pause_detection(self, *_, **__):
        """
        Put the assistant on pause. No new conversation events will be triggered.
        """
        self._detection_paused.set()

    @action
    def resume_detection(self, *_, **__):
        """
        Resume the assistant hotword detection from a paused state.
        """
        self._detection_paused.clear()

    @action
    def is_detecting(self, *_, **__) -> bool:
        """
        :return: True if the asistant is detecting, False otherwise.
        """
        return not self._detection_paused.is_set()

    @action
    def is_muted(self, *_, **__) -> bool:
        """
        :return: True if the microphone is muted, False otherwise.
        """
        return self._is_muted

    @action
    def status(self, *_, **__):
        """
        :return: The current assistant status:

            .. code-block:: json

                {
                    "last_query": "What time is it?",
                    "last_response": "It's 10:30 AM",
                    "conversation_running": true,
                    "is_muted": false,
                    "is_detecting": true
                }

        """
        self.publish_entities([self])
        return asdict(self._state)

    @action
    def render_response(
        self, text: str, *_, with_follow_on_turn: Optional[bool] = None, **__
    ) -> bool:
        """
        Render a response text as audio over the configured TTS plugin.

        :param text: Text to render.
        :param with_follow_on_turn: If set, the assistant will wait for a follow-up.
            By default, ``with_follow_on_turn`` will be automatically set to true if
            the ``text`` ends with a question mark.
        :return: True if the assistant is waiting for a follow-up, False otherwise.
        """
        if not text:
            self._on_no_response()
            return False

        follow_up = (
            bool(text and text.strip().endswith('?'))
            if with_follow_on_turn is None
            else with_follow_on_turn
        )

        self._on_response_render_start(text, with_follow_on_turn=follow_up)
        self._render_response(text)
        self._on_response_render_end(with_follow_on_turn=follow_up)

        if follow_up:
            self.start_conversation()
        else:
            self.stop_conversation()

        return follow_up

    def _get_tts_plugin(self):
        if not self.tts_plugin:
            return None

        return get_plugin(self.tts_plugin)

    def _play_conversation_start_sound(self):
        if not self._conversation_start_sound:
            return

        audio = get_plugin('sound')
        if not audio:
            self.logger.warning(
                'Unable to play conversation start sound: sound plugin not found'
            )
            return

        audio.play(self._conversation_start_sound)

    def _send_event(self, event_type: Type[AppEvent], **kwargs):
        self.publish_entities([self])
        get_bus().post(event_type(assistant=self._plugin_name, **kwargs))

    def _on_conversation_start(self):
        from platypush.message.event.assistant import ConversationStartEvent

        self._last_response = None
        self._last_query = None
        self._conversation_running.set()
        self._send_event(ConversationStartEvent)
        self._play_conversation_start_sound()

    def _on_conversation_end(self):
        from platypush.message.event.assistant import ConversationEndEvent

        self._conversation_running.clear()
        self._send_event(ConversationEndEvent)

    def _on_conversation_timeout(self):
        from platypush.message.event.assistant import ConversationTimeoutEvent

        self._last_response = None
        self._last_query = None
        self._conversation_running.clear()
        self._send_event(ConversationTimeoutEvent)

    def _on_no_response(self):
        from platypush.message.event.assistant import NoResponseEvent

        self._last_response = None
        self._conversation_running.clear()
        self._send_event(NoResponseEvent)

    def _on_response_render_start(
        self, text: Optional[str], with_follow_on_turn: bool = False
    ):
        from platypush.message.event.assistant import ResponseEvent

        self._last_response = text
        self._send_event(
            ResponseEvent, response_text=text, with_follow_on_turn=with_follow_on_turn
        )

    def _render_response(self, text: Optional[str]):
        if not text:
            return

        tts = self._get_tts_plugin()
        if not tts:
            self.logger.warning(
                'Got a response to render, but no TTS plugin is configured: %s', text
            )
            return

        tts.say(text=text, **self.tts_plugin_args)

    def _on_response_render_end(self, with_follow_on_turn: bool = False):
        from platypush.message.event.assistant import ResponseEndEvent

        self._send_event(
            ResponseEndEvent,
            response_text=self._last_response,
            with_follow_on_turn=with_follow_on_turn,
        )

    def _on_hotword_detected(self, hotword: Optional[str]):
        from platypush.message.event.assistant import HotwordDetectedEvent

        self._send_event(HotwordDetectedEvent, hotword=hotword)

    def _on_speech_recognized(self, phrase: Optional[str]):
        from platypush.message.event.assistant import SpeechRecognizedEvent

        phrase = (phrase or '').lower().strip()
        self._last_query = phrase
        self._send_event(SpeechRecognizedEvent, phrase=phrase)

    def _on_intent_matched(self, intent: str, slots: Optional[Dict[str, Any]] = None):
        from platypush.message.event.assistant import IntentRecognizedEvent

        self._send_event(IntentRecognizedEvent, intent=intent, slots=slots)

    def _on_alarm_start(self):
        from platypush.message.event.assistant import AlarmStartedEvent

        self._cur_alert_type = AlertType.ALARM
        self._send_event(AlarmStartedEvent)

    def _on_alarm_end(self):
        from platypush.message.event.assistant import AlarmEndEvent

        self._cur_alert_type = None
        self._send_event(AlarmEndEvent)

    def _on_timer_start(self):
        from platypush.message.event.assistant import TimerStartedEvent

        self._cur_alert_type = AlertType.TIMER
        self._send_event(TimerStartedEvent)

    def _on_timer_end(self):
        from platypush.message.event.assistant import TimerEndEvent

        self._cur_alert_type = None
        self._send_event(TimerEndEvent)

    def _on_alert_start(self):
        from platypush.message.event.assistant import AlertStartedEvent

        self._cur_alert_type = AlertType.ALERT
        self._send_event(AlertStartedEvent)

    def _on_alert_end(self):
        from platypush.message.event.assistant import AlertEndEvent

        self._cur_alert_type = None
        self._send_event(AlertEndEvent)

    def _on_mute(self):
        from platypush.message.event.assistant import MicMutedEvent

        self._is_muted = True
        self._send_event(MicMutedEvent)

    def _on_unmute(self):
        from platypush.message.event.assistant import MicUnmutedEvent

        self._is_muted = False
        self._send_event(MicUnmutedEvent)

    def _on_mute_changed(self, value: bool):
        if value:
            self._on_mute()
        else:
            self._on_unmute()

    def transform_entities(self, entities: Collection['AssistantPlugin'], **_):
        return super().transform_entities(
            [
                Assistant(
                    external_id=get_plugin_name_by_class(type(dev)),
                    name=self._entity_name,
                    last_query=dev._state.last_query,
                    last_response=dev._state.last_response,
                    conversation_running=dev._state.conversation_running,
                    is_muted=dev._state.is_muted,
                    is_detecting=dev._state.is_detecting,
                )
                for dev in (entities or [])
            ]
        )


# vim:sw=4:ts=4:et:
